# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated, Any

from fastapi import APIRouter, Depends, HTTPException

from pixano.app.models import DatasetBrowser, PaginationColumn, PaginationInfo, TableData
from pixano.app.settings import Settings, get_settings
from pixano.datasets.utils.errors import DatasetAccessError
from pixano.features import SchemaGroup, is_image, is_text, is_view_embedding
from pixano.features.utils.image import generate_text_image_base64, get_image_thumbnail, image_to_base64

from .utils import get_dataset, get_rows


router = APIRouter(prefix="/browser", tags=["Browser"])


@router.get("/{id}", response_model=DatasetBrowser)
async def get_browser(
    id: str,
    settings: Annotated[Settings, Depends(get_settings)],
    limit: int = 50,
    skip: int = 0,
    query: str = "",
    embedding_table: str = "",
    where: str | None = None,
) -> DatasetBrowser:  # type: ignore
    """Load dataset items for the explorer page.

    Args:
        id: Dataset ID containing the items.
        settings: App settings.
        limit: Limit number of items.
        skip: Skip number of items.
        query: Text query for semantic search.
        embedding_table: Table name for embeddings.
        where: Where clause.

    Returns:
        Dataset explorer page.
    """
    # Load dataset
    dataset = get_dataset(id, settings.library_dir, settings.media_dir)

    semantic_search = embedding_table != ""
    if query != "" or embedding_table != "":
        if query == "" or embedding_table == "":
            raise HTTPException(
                status_code=400,
                detail="Both query and model_name should be provided for semantic search.",
            )

    # Get page parameters
    total = dataset.num_rows
    original_limit = limit
    limit = min(limit, total)

    # Get tables
    table_item = SchemaGroup.ITEM.value
    tables_view = sorted(dataset.schema.groups[SchemaGroup.VIEW])

    # get data (items and views)
    if semantic_search:
        try:
            item_rows, distances = dataset.semantic_search(query, embedding_table, limit, skip)
        except DatasetAccessError as e:
            raise HTTPException(status_code=400, detail=str(e))
    else:
        item_rows = get_rows(dataset=dataset, table=table_item, limit=limit, skip=skip, where=where)

    item_ids = [item.id for item in item_rows]
    item_first_media: dict[str, dict] = {}
    for view in tables_view:
        try:
            view_rows = get_rows(dataset=dataset, table=view, item_ids=item_ids)
        except HTTPException:
            view_rows = []
        item_first_media[view] = {}
        for item_id in item_ids:
            # store image or first frame
            item_first_media[view][item_id] = next(filter(lambda v: v.item_ref.id == item_id, view_rows), None)

    # build column headers (PaginationColumn)
    cols = []
    for view in tables_view:
        view_type = "image"
        # NOTE: right now for video we use a frame
        # (first returned by get_rows for each item. May not be the real first frame)
        # when we'll have thumbnail clip, use instead:
        # view_type = "video" if isinstance(item_first_media[view][item_ids[0]], SequenceFrame) else "image"
        cols.append(PaginationColumn(name=view, type=view_type))
    for feat in vars(item_rows[0]).keys():
        cols.append(PaginationColumn(name=feat, type=type(getattr(item_rows[0], feat)).__name__))
    if semantic_search:
        cols.append(PaginationColumn(name="distance", type="float"))

    # build rows
    rows: list[dict[str, Any]] = []
    for i, item in enumerate(item_rows):
        row: dict[str, Any] = {}
        # VIEWS -> thumbnails previews
        for view in tables_view:
            curr_view = item_first_media[view][item.id]
            if curr_view is not None:
                if is_image(type(curr_view)):
                    try:
                        row_view = curr_view.open(settings.media_dir, output_type="image")
                        row_view = get_image_thumbnail(row_view, (128, 128))
                        row_view_base64 = image_to_base64(row_view, curr_view.format)
                    except ValueError:
                        row_view_base64 = ""
                elif is_text(type(curr_view)):
                    row_view_base64 = generate_text_image_base64(curr_view.content[:80])
                row[view] = row_view_base64

        # ITEM features
        for feat in vars(item).keys():
            row[feat] = getattr(item, feat)
        # DISTANCE
        if semantic_search:
            row["distance"] = distances[i]

        rows.append(row)

    semantic_search_list = sorted(
        [
            table_name
            for table_name in dataset.schema.groups[SchemaGroup.EMBEDDING]
            if is_view_embedding(dataset.schema.schemas[table_name])
        ]
    )
    # Return dataset items
    return DatasetBrowser(
        id=id,
        name=dataset.info.name,
        table_data=TableData(columns=cols, rows=rows),
        pagination=PaginationInfo(current_page=skip, page_size=original_limit, total_size=total),
        semantic_search=semantic_search_list,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import FileResponse

from pixano.app.settings import Settings, get_settings


router = APIRouter(prefix="/models", tags=["Models"])


@router.get("/", response_model=list[str])
async def get_models(
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[str]:
    """Get all models in the models directory with the extension `.onnx`.

    Args:
        settings: App settings.

    Returns:
        List of models.
    """
    if settings.models_dir is None:
        raise HTTPException(status_code=500, detail="Models directory not set")
    models = [file.stem for file in settings.models_dir.glob("*.onnx")]
    return models


@router.get("/{model_name}", response_class=FileResponse)
async def get_model(
    model_name: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> FileResponse:
    """Get model file by name.

    The model file is expected to be in the models directory with the extension `.onnx`.

    Args:
        model_name: Model name.
        settings: App settings.

    Returns:
        Model file.
    """
    if settings.models_dir is None:
        raise HTTPException(status_code=500, detail="Models directory not set")
    model_path = settings.models_dir / f"{model_name}.onnx"
    if not model_path.exists():
        raise HTTPException(status_code=404, detail=f"Model {model_name} not found in {settings.models_dir}")
    return FileResponse(model_path)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path
from typing import Any

import shortuuid
from fastapi.encoders import jsonable_encoder
from pixano_inference.client import PixanoInferenceClient
from pixano_inference.pydantic.tasks import (
    TextImageConditionalGenerationRequest,
    TextImageConditionalGenerationResponse,
)
from pixano_inference.utils import is_url

from pixano.datasets import Dataset
from pixano.features import Conversation, EntityRef, Message, SchemaGroup, Source, SourceRef, is_bbox


DEFAULT_ROLE_SYSTEM = "system"
DEFAULT_ROLE_USER = "user"
DEFAULT_ROLE_ASSISTANT = "assistant"
DEFAULT_TEMPERATURE = 0.9
DEFAULT_MAX_NEW_TOKENS = 100


def messages_to_prompt(
    dataset: Dataset,
    conversation: Conversation,
    messages: list[Message],
    media_dir: Path,
    role_system: str = DEFAULT_ROLE_SYSTEM,
    role_user: str = DEFAULT_ROLE_USER,
    role_assistant: str = DEFAULT_ROLE_ASSISTANT,
) -> list[dict[str, Any]]:
    """Convert a list of messages to a prompt.

    Args:
        dataset: The Pixano dtaset.
        conversation: The conversation entity of the messages.
        messages: List of messages.
        media_dir: The directory containing the images.
        role_system: The role to use for the system.
        role_user: The role to use for the user.
        role_assistant: The role to use for the assistant.

    Returns:
        List of dictionaries representing the prompt.
    """
    prompt = []
    for message in messages:
        message_prompt: dict[str, Any] = {"content": []}
        match message.type:
            case "SYSTEM":
                message_prompt["role"] = role_system
            case "QUESTION":
                message_prompt["role"] = role_user
            case "ANSWER":
                message_prompt["role"] = role_assistant
            case _:
                raise ValueError(f"Unknown message type {message.type}")

        ## add images to prompt
        tables_view = sorted(dataset.schema.groups[SchemaGroup.VIEW])
        for table_view in tables_view:
            images = dataset.get_data(table_view, item_ids=[conversation.item_ref.id])
            if len(images) > 0:
                image = images[0]  # there should be only one image per view, except for video (out of scope now)
                message_prompt["content"].append(
                    {
                        "type": "image_url",
                        "image_url": {"url": image.url if is_url(image.url) else image.open(media_dir, "base64")},
                    }
                )

        ## add objects (bbox) to prompt
        tables_entities = sorted(dataset.schema.groups[SchemaGroup.ENTITY])
        tables_bbox = [k for k, v in dataset.schema.schemas.items() if is_bbox(v)]
        if len(tables_bbox) > 0:
            table_bbox = tables_bbox[0]  # assume there is only one bbox table
            for table_entity in tables_entities:
                entities = dataset.get_data(table_entity, item_ids=[conversation.item_ref.id])
                for entity in entities:
                    if not isinstance(entity, Conversation):
                        bboxes = dataset.get_data(
                            table_bbox, item_ids=[conversation.item_ref.id], where=f"entity_ref.id == '{entity.id}'"
                        )
                        if len(bboxes) == 0:
                            continue
                        bbox = bboxes[0]  # assume only one bbox per entity
                        bbox_text = f"a {'normalized ' if bbox.is_normalized else ''}bounding box {bbox.coords}"
                        if hasattr(entity, "name"):
                            bbox_text = bbox_text + f" with the name '{entity.name}'"
                        message_prompt["content"].append(
                            {
                                "type": "text",
                                "text": bbox_text,
                            }
                        )

        message_prompt["content"].append({"type": "text", "text": message.content})
        prompt.append(message_prompt)
    return prompt


async def text_image_conditional_generation(
    client: PixanoInferenceClient,
    source: Source,
    dataset: Dataset,
    media_dir: Path,
    messages: list[Message],
    conversation: Conversation,
    max_new_tokens: int = DEFAULT_MAX_NEW_TOKENS,
    temperature: float = DEFAULT_TEMPERATURE,
    role_system: str = DEFAULT_ROLE_SYSTEM,
    role_user: str = DEFAULT_ROLE_USER,
    role_assistant: str = DEFAULT_ROLE_ASSISTANT,
    **client_kwargs: Any,
) -> Message:
    """Generate text from an image using the Pixano Inference API.

    Args:
        client: The Pixano-Inference client to use.
        source: The source refering to the model.
        dataset: The Pixano dataset.
        media_dir: The directory containing the input media files.
        messages: A list of Message objects representing the input messages.
        conversation: The conversation entity of the messages.
        max_new_tokens: The maximum number of tokens to generate.
        temperature: The temperature to use for sampling.
        role_system: The role of the system in the prompt.
        role_user: The role of the user in the prompt.
        role_assistant: The role of the assistant in the prompt.
        client_kwargs: Additional kwargs for the client to be passed.

    Returns:
        The response message and its source.
    """
    prompt = messages_to_prompt(
        dataset=dataset,
        conversation=conversation,
        messages=messages,
        media_dir=media_dir,
        role_system=role_system,
        role_user=role_user,
        role_assistant=role_assistant,
    )
    last_message = messages[-1]
    match last_message.type:
        case "SYSTEM":
            response_type = "QUESTION"
            number = last_message.number + 1
        case "ANSWER":
            response_type = "QUESTION"
            number = last_message.number + 1
        case "QUESTION":
            response_type = "ANSWER"
            number = last_message.number
        case _:
            raise ValueError(f"Invalid last message type {last_message.type}")
    request = TextImageConditionalGenerationRequest(
        model=source.name,
        prompt=prompt,
        images=None,
        max_new_tokens=max_new_tokens,
        temperature=temperature,
    )
    response: TextImageConditionalGenerationResponse = await client.text_image_conditional_generation(
        request, **client_kwargs
    )

    inference_metadata = jsonable_encoder(
        {
            "generation_config": response.data.generation_config,
            "usage": response.data.usage,
            "timestamp": response.timestamp,
            "processing_time": response.processing_time,
        }
    )

    message = Message(
        id=shortuuid.uuid(),
        item_ref=last_message.item_ref,
        view_ref=last_message.view_ref,
        entity_ref=EntityRef(id=conversation.id, name=conversation.table_name),
        source_ref=SourceRef(id=source.id),
        type=response_type,
        content=response.data.generated_text,
        number=number,
        user=source.name,
        timestamp=response.timestamp,
        inference_metadata=inference_metadata,
    )

    return message



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

CATEGORY_NAMES = {
    "coco91": {
        1: "person",
        2: "bicycle",
        3: "car",
        4: "motorcycle",
        5: "airplane",
        6: "bus",
        7: "train",
        8: "truck",
        9: "boat",
        10: "traffic light",
        11: "fire hydrant",
        12: "street sign",
        13: "stop sign",
        14: "parking meter",
        15: "bench",
        16: "bird",
        17: "cat",
        18: "dog",
        19: "horse",
        20: "sheep",
        21: "cow",
        22: "elephant",
        23: "bear",
        24: "zebra",
        25: "giraffe",
        26: "hat",
        27: "backpack",
        28: "umbrella",
        29: "shoe",
        30: "eye glasses",
        31: "handbag",
        32: "tie",
        33: "suitcase",
        34: "frisbee",
        35: "skis",
        36: "snowboard",
        37: "sports ball",
        38: "kite",
        39: "baseball bat",
        40: "baseball glove",
        41: "skateboard",
        42: "surfboard",
        43: "tennis racket",
        44: "bottle",
        45: "plate",
        46: "wine glass",
        47: "cup",
        48: "fork",
        49: "knife",
        50: "spoon",
        51: "bowl",
        52: "banana",
        53: "apple",
        54: "sandwich",
        55: "orange",
        56: "broccoli",
        57: "carrot",
        58: "hot dog",
        59: "pizza",
        60: "donut",
        61: "cake",
        62: "chair",
        63: "couch",
        64: "potted plant",
        65: "bed",
        66: "mirror",
        67: "dining table",
        68: "window",
        69: "desk",
        70: "toilet",
        71: "door",
        72: "tv",
        73: "laptop",
        74: "mouse",
        75: "remote",
        76: "keyboard",
        77: "cell phone",
        78: "microwave",
        79: "oven",
        80: "toaster",
        81: "sink",
        82: "refrigerator",
        83: "blender",
        84: "book",
        85: "clock",
        86: "vase",
        87: "scissors",
        88: "teddy bear",
        89: "hair drier",
        90: "toothbrush",
        91: "hair brush",
    },
    "coco80": {
        1: "person",
        2: "bicycle",
        3: "car",
        4: "motorcycle",
        5: "airplane",
        6: "bus",
        7: "train",
        8: "truck",
        9: "boat",
        10: "traffic light",
        11: "fire hydrant",
        12: "stop sign",
        13: "parking meter",
        14: "bench",
        15: "bird",
        16: "cat",
        17: "dog",
        18: "horse",
        19: "sheep",
        20: "cow",
        21: "elephant",
        22: "bear",
        23: "zebra",
        24: "giraffe",
        25: "backpack",
        26: "umbrella",
        27: "handbag",
        28: "tie",
        29: "suitcase",
        30: "frisbee",
        31: "skis",
        32: "snowboard",
        33: "sports ball",
        34: "kite",
        35: "baseball bat",
        36: "baseball glove",
        37: "skateboard",
        38: "surfboard",
        39: "tennis racket",
        40: "bottle",
        41: "wine glass",
        42: "cup",
        43: "fork",
        44: "knife",
        45: "spoon",
        46: "bowl",
        47: "banana",
        48: "apple",
        49: "sandwich",
        50: "orange",
        51: "broccoli",
        52: "carrot",
        53: "hot dog",
        54: "pizza",
        55: "donut",
        56: "cake",
        57: "chair",
        58: "couch",
        59: "potted plant",
        60: "bed",
        61: "dining table",
        62: "toilet",
        63: "tv",
        64: "laptop",
        65: "mouse",
        66: "remote",
        67: "keyboard",
        68: "cell phone",
        69: "microwave",
        70: "oven",
        71: "toaster",
        72: "sink",
        73: "refrigerator",
        74: "book",
        75: "clock",
        76: "vase",
        77: "scissors",
        78: "teddy bear",
        79: "hair drier",
        80: "toothbrush",
    },
    "voc": {
        1: "aeroplane",
        2: "bicycle",
        3: "bird",
        4: "boat",
        5: "bottle",
        6: "bus",
        7: "car",
        8: "cat",
        9: "chair",
        10: "cow",
        11: "dining table",
        12: "dog",
        13: "horse",
        14: "motorbike",
        15: "person",
        16: "potted plant",
        17: "sheep",
        18: "sofa",
        19: "train",
        20: "tv / monitor",
    },
}
CATEGORY_IDS = {key: {v: k for k, v in value.items()} for key, value in CATEGORY_NAMES.items()}


def category_name(category_id: int, category_dict: str = "coco91") -> str:
    """Return category name based on category ID.

    Args:
        category_id: Category ID
        category_dict: Which category dictionary to use ("coco91", "coco80", "voc")

    Returns:
        Category name
    """
    return CATEGORY_NAMES[category_dict][int(category_id)]


def category_id(category_name: str, category_dict: str = "coco91") -> int:
    """Return category ID based on category name.

    Args:
        category_name: Category name
        category_dict: Which category dictionary to use ("coco91", "coco80", "voc")

    Returns:
        Category ID
    """
    return CATEGORY_IDS[category_dict][str(category_name).strip().lower()]


def coco_ids_80to91(category_id: int) -> int:
    """Convert COCO category ID from 80 to 91 classes.

    Args:
        category_id: Category ID (80 classes)

    Returns:
        Category ID (91 classes)
    """
    coco_dict = {
        1: 1,
        2: 2,
        3: 3,
        4: 4,
        5: 5,
        6: 6,
        7: 7,
        8: 8,
        9: 9,
        10: 10,
        11: 11,
        12: 13,
        13: 14,
        14: 15,
        15: 16,
        16: 17,
        17: 18,
        18: 19,
        19: 20,
        20: 21,
        21: 22,
        22: 23,
        23: 24,
        24: 25,
        25: 27,
        26: 28,
        27: 31,
        28: 32,
        29: 33,
        30: 34,
        31: 35,
        32: 36,
        33: 37,
        34: 38,
        35: 39,
        36: 40,
        37: 41,
        38: 42,
        39: 43,
        40: 44,
        41: 46,
        42: 47,
        43: 48,
        44: 49,
        45: 50,
        46: 51,
        47: 52,
        48: 53,
        49: 54,
        50: 55,
        51: 56,
        52: 57,
        53: 58,
        54: 59,
        55: 60,
        56: 61,
        57: 62,
        58: 63,
        59: 64,
        60: 65,
        61: 67,
        62: 70,
        63: 72,
        64: 73,
        65: 74,
        66: 75,
        67: 76,
        68: 77,
        69: 78,
        70: 79,
        71: 80,
        72: 81,
        73: 82,
        74: 84,
        75: 85,
        76: 86,
        77: 87,
        78: 88,
        79: 89,
        80: 90,
    }

    return coco_dict[int(category_id)]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano_inference.client import PixanoInferenceClient

from .mask_generation import image_mask_generation
from .text_image_conditional_generation import messages_to_prompt, text_image_conditional_generation
from .zero_shot_detection import image_zero_shot_detection


__all__ = [
    "PixanoInferenceClient",
    "image_mask_generation",
    "messages_to_prompt",
    "text_image_conditional_generation",
    "image_zero_shot_detection",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException
from pixano_inference.client import PixanoInferenceClient

from pixano.app.settings import Settings, get_settings

from . import conditional_generation, mask_generation, models, zero_shot_detection


router = APIRouter(prefix="/inference", tags=["Inference"])
router.include_router(conditional_generation.router)
router.include_router(models.router)
router.include_router(zero_shot_detection.router)
router.include_router(mask_generation.router)


@router.post("/connect")
async def connect_inference(url: str, settings: Annotated[Settings, Depends(get_settings)]) -> None:
    """Connect to Pixano Inference."""
    try:
        client = PixanoInferenceClient.connect(url)
    except Exception:
        raise HTTPException(status_code=404, detail=f"Impossible to connect to Pixano Inference from url: {url}")

    settings.pixano_inference_client = client
    return


__all__ = ["router"]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path
from typing import Iterable


def create_video_preview(path: Path, frame_urls: Iterable[str], fps: int = 25, scale: float = 0.5):
    """Create a video preview by writing a sequence of frames to a video file.

    Args:
        path: The path to the output video file.
        frame_urls: URLs pointing to the frames of the video.
        fps: The frames per second of the output video.
        scale: The scale factor to resize the frames.
    """
    # Import mediapy only when needed to avoid unnecessary dependencies.
    import mediapy

    frames = [mediapy.read_image(url) for url in frame_urls]

    if scale < 1:
        size = frames[0].shape[:2]
        target_size = (int(size[0] * scale), int(size[1] * scale))
        frames = mediapy.resize_video(frames, target_size)

    mediapy.write_video(path, frames, fps=fps)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from __future__ import annotations

from datetime import datetime
from types import GenericAlias
from typing import TYPE_CHECKING, Any, overload

from lancedb.pydantic import FixedSizeListMixin, LanceModel, Vector
from pydantic import ConfigDict, PrivateAttr, create_model, field_validator
from typing_extensions import Self

from pixano.utils import get_super_type_from_dict, issubclass_strict
from pixano.utils.validation import validate_and_init_create_at_and_update_at

from ..pyarrow_utils import DESERIALIZE_PYARROW_DATATYPE, SERIALIZE_PYARROW_DATATYPE
from ..types.registry import _TYPES_REGISTRY


if TYPE_CHECKING:
    from pixano.datasets.dataset import Dataset
    from pixano.features import (
        Annotation,
        AnnotationRef,
        Embedding,
        EmbeddingRef,
        Entity,
        EntityRef,
        Item,
        ItemRef,
        SchemaRef,
        Source,
        SourceRef,
        View,
        ViewRef,
    )


class BaseSchema(LanceModel):
    """Base class for all schemas.

    All schemas should inherit from this class and therefore all elements in the dataset contains an id.

    Attributes:
        id: the id of the manipulated object.
        created_at: the creation date of the object.
        updated_at: the last modification date of the object.

    Note:
        If the `created_at` and `updated_at` fields are not provided, they are set to the current date and time.
    """

    model_config = ConfigDict(validate_assignment=True)
    id: str = ""
    created_at: datetime
    updated_at: datetime
    _dataset: Dataset | None = PrivateAttr(None)
    _table_name: str = PrivateAttr("")

    @field_validator("id", mode="after")
    @classmethod
    def _id_validator(cls, v: str) -> str:
        if " " in v:
            raise ValueError("id must not contain spaces")
        return v

    def __init__(self, /, created_at: datetime | None = None, updated_at: datetime | None = None, **data: Any):
        """Create a new model by parsing and validating input data from keyword arguments.

        Raises [`ValidationError`][pydantic_core.ValidationError] if the input data cannot be
        validated to form a valid model.

        `self` is explicitly positional-only to allow `self` as a field name.

        Args:
            created_at: The creation date of the object.
            updated_at: The last modification date of the object.
            data: The data of the object validated by Pydantic.
        """
        created_at, updated_at = validate_and_init_create_at_and_update_at(created_at, updated_at)
        data.update({"created_at": created_at, "updated_at": updated_at})
        super().__init__(**data)

    @property
    def dataset(self) -> Dataset:
        """Get the dataset."""
        if self._dataset is None:
            raise ValueError("Dataset is not set.")
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: Dataset):
        """Set the dataset."""
        self._dataset = dataset

    @property
    def table_name(self) -> str:
        """Get the table name."""
        if self._table_name == "":
            raise ValueError("Table name is not set.")
        return self._table_name

    @table_name.setter
    def table_name(self, table_name: str):
        """Set the table name."""
        self._table_name = table_name

    def model_dump(self, exclude_timestamps: bool = False, **kwargs: Any) -> dict[str, Any]:
        """Dump the model to a dictionary.

        Args:
            exclude_timestamps: Exclude timestamps "created_at" and "updated_at" from the model dump. Useful for
                comparing models without timestamps.
            kwargs: Arguments for pydantic `BaseModel.model_dump()`.

        Returns:
            The model dump.
        """
        model_dump = super().model_dump(**kwargs)
        if exclude_timestamps:
            model_dump.pop("created_at", None)
            model_dump.pop("updated_at", None)
        return model_dump

    def model_copy(self, *, update: dict[str, Any] | None = None, deep: bool = False) -> Self:
        """Returns a copy of the model.

        Args:
            update: Values to change/add in the new model.
            deep: Set to `True` to make a deep copy of the model.

        Returns:
            New model instance.
        """
        # Wrap the pydantic `model_copy` method to prevent copying the dataset.
        dataset = self._dataset
        self._dataset = None

        copy = super().model_copy(update=update, deep=deep)
        copy.dataset = dataset
        return copy

    @overload
    def resolve_ref(self, ref: "ItemRef") -> "Item": ...
    @overload
    def resolve_ref(self, ref: "ViewRef") -> "View": ...
    @overload
    def resolve_ref(self, ref: "EmbeddingRef") -> "Embedding": ...
    @overload
    def resolve_ref(self, ref: "EntityRef") -> "Entity": ...
    @overload
    def resolve_ref(self, ref: "AnnotationRef") -> "Annotation": ...
    @overload
    def resolve_ref(self, ref: "SourceRef") -> "Source": ...
    @overload
    def resolve_ref(self, ref: "SchemaRef") -> "BaseSchema": ...
    def resolve_ref(
        self, ref: "SchemaRef" | "ItemRef" | "ViewRef" | "EmbeddingRef" | "EntityRef" | "AnnotationRef" | "SourceRef"
    ) -> "BaseSchema" | "Item" | "View" | "Embedding" | "Entity" | "Annotation" | "Source":
        """Resolve a reference to a schema object in the dataset.

        Args:
            ref: The reference to resolve.

        Returns:
            The resolved schema object.
        """
        return self.dataset.resolve_ref(ref)

    @classmethod
    def serialize(cls) -> dict[str, str | dict[str, Any]]:
        """Serialize the table."""
        from .registry import _PIXANO_SCHEMA_REGISTRY

        # schema can be customized by the user
        # base_schema is the closest schema in the registry
        super_type = get_super_type_from_dict(cls, _PIXANO_SCHEMA_REGISTRY)
        if super_type is None:
            raise ValueError(f"Schema {cls.__name__} does not have a super type in the registry.")
        json: dict[str, str | dict[str, Any]] = {
            "schema": cls.__name__,
            "base_schema": super_type.__name__,
        }
        fields: dict[str, Any] = {}
        for field_name, field in cls.model_fields.items():
            if isinstance(field.annotation, GenericAlias):
                origin = field.annotation.__origin__
                args = field.annotation.__args__

                if origin in [list, tuple]:
                    if issubclass(args[0], tuple(_TYPES_REGISTRY.values())):
                        fields[field_name] = {
                            "type": args[0].__name__,
                            "collection": True,
                        }
                    else:
                        fields[field_name] = {
                            "type": args[0].__name__,
                            "collection": True,
                        }
                else:
                    raise NotImplementedError("Should be a list or tuple.")
            else:
                if issubclass(field.annotation, tuple(_TYPES_REGISTRY.values())):
                    fields[field_name] = {
                        "type": field.annotation.__name__,
                        "collection": False,
                    }
                elif issubclass(field.annotation, FixedSizeListMixin):  # LanceDB Vector
                    fields[field_name] = {
                        "type": field.annotation.__name__,
                        "collection": False,
                        "dim": field.annotation.dim(),
                        "value_type": SERIALIZE_PYARROW_DATATYPE[field.annotation.value_arrow_type()],
                    }
                else:
                    fields[field_name] = {
                        "type": field.annotation.__name__,
                        "collection": False,
                    }
        json["fields"] = fields
        return json

    @staticmethod
    def deserialize(dataset_schema_json: dict[str, str | dict[str, Any]]) -> type["BaseSchema"]:
        """Deserialize the dataset schema.

        Args:
            dataset_schema_json: Serialized dataset schema.

        Returns:
            The dataset schema.
        """
        from .registry import _PIXANO_SCHEMA_REGISTRY, _SCHEMA_REGISTRY

        json_fields = dataset_schema_json["fields"]
        if not isinstance(json_fields, dict):
            raise ValueError("Fields should be a dictionary.")

        fields: dict[str, Any] = {}
        for key, value in json_fields.items():
            if value["type"] in _TYPES_REGISTRY:
                type_ = _TYPES_REGISTRY[value["type"]]
            elif value["type"] == "FixedSizeList":  # LanceDB Vector
                type_ = value["type"]
                dim = value["dim"]
                value_type = DESERIALIZE_PYARROW_DATATYPE[value["value_type"]]
                type_ = Vector(dim, value_type)
            else:
                raise ValueError(f"Type {value['type']} not registered")
            if value["collection"]:
                type_ = list[type_]  # type: ignore[valid-type]
            fields[key] = (type_, ...)

        schema, base_schema = dataset_schema_json["schema"], dataset_schema_json["base_schema"]

        if not isinstance(schema, str) or not isinstance(base_schema, str):
            raise ValueError("Schema and base schema should be strings.")

        if schema in _SCHEMA_REGISTRY:
            table_type = _SCHEMA_REGISTRY[schema]
        else:
            table_type = _PIXANO_SCHEMA_REGISTRY[base_schema]

        model = create_model(dataset_schema_json["schema"], **fields, __base__=table_type)

        return model


def is_base_schema(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `BaseSchema` or subclass of `BaseSchema`."""
    return issubclass_strict(cls, BaseSchema, strict)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .coco_dataset_exporter import COCODatasetExporter
from .dataset_exporter import DatasetExporter
from .default_json_dataset_exporter import DefaultJSONDatasetExporter
from .default_jsonl_dataset_exporter import DefaultJSONLDatasetExporter


__all__ = [
    "DatasetExporter",
    "DefaultJSONDatasetExporter",
    "DefaultJSONLDatasetExporter",
    "COCODatasetExporter",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


import json
from datetime import datetime
from pathlib import Path
from typing import Any

from fastapi.encoders import jsonable_encoder

from pixano.datasets.utils.labels import CATEGORY_IDS
from pixano.features import (
    BaseSchema,
    SchemaGroup,
    Source,
    schema_to_group,
)
from pixano.features.schemas import BBox, CompressedRLE, Entity, Image, SequenceFrame

from ..dataset import Dataset
from ..dataset_info import DatasetInfo
from ..dataset_schema import DatasetItem
from .dataset_exporter import DatasetExporter


class COCODatasetExporter(DatasetExporter):
    """Default JSON dataset exporter."""

    def __init__(
        self,
        dataset: Dataset,
        export_dir: str | Path,
        overwrite: bool = False,
        category_format: str = "coco91",
        custom_category_dict: dict[str, int] | None = None,
    ):
        """Initialize a new instance of the DatasetExporter class.

        Args:
            dataset: The dataset to be exported.
            export_dir: The directory where the exported files will be saved.
            overwrite: Whether to overwrite existing directory.
            category_format: Category format for name to ID conversion ("coco91", "coco80", "voc").
            custom_category_dict: Custom category dictionary for name to ID conversion (supersedes category_format).
        """
        self.dataset = dataset
        self.export_dir = Path(export_dir)
        self._overwrite = overwrite
        self.category_dict = (
            custom_category_dict if custom_category_dict is not None else CATEGORY_IDS[category_format]
        )

    def initialize_export_data(self, info: DatasetInfo, sources: list[Source]) -> dict[str, Any]:
        """Initialize the dictionary or list of dictionaries to be exported.

        Args:
            info: The dataset information.
            sources: The list of sources.

        Returns:
            A dictionary containing the data to be exported.
        """
        export_data = {
            "info": {
                "id": info.id,
                "name": info.name,
                "year": datetime.now().year,
                "version": datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
                "description": info.description,
                "contributor": "",
                "url": "",
                "date_created": datetime.now(),
            },
            "licenses": [
                {
                    "id": 0,
                    "name": "Exported from Pixano",
                    "url": "",
                }
            ],
            "images": [],
            "annotations": [],
            "categories": [],
        }
        return export_data

    def export_dataset_item(self, export_data: dict[str, Any], dataset_item: DatasetItem) -> dict[str, Any]:
        """Store the dataset item in the `export_data` dictionary.

        Args:
            export_data: A dictionary containing the data to be exported.
            dataset_item: The dataset item to be exported.

        Returns:
            A dictionary containing the data to be exported.
        """
        data: dict[str, BaseSchema | list[BaseSchema] | None] = dataset_item.to_schemas_data(self.dataset.schema)
        # Keep annotations in a dictionary to merge BBox, CompressedRLE, and Category before adding to export_data
        anns = {s["id"]: s for s in export_data["annotations"]}
        for schema_name, schema_data in data.items():
            if schema_data:
                schema_data = schema_data if isinstance(schema_data, list) else [schema_data]
                group = schema_to_group(schema_data[0])
                if group == SchemaGroup.VIEW:
                    export_data["images"].extend(
                        [coco_image(s, schema_name) for s in schema_data if isinstance(s, Image)]
                    )
                elif group == SchemaGroup.ENTITY:
                    for schema in schema_data:
                        if isinstance(schema, Entity) and hasattr(schema, "category"):
                            ann_id = f"{schema.view_ref.id}_{schema.id}"
                            anns[ann_id] = coco_annotation(
                                ann=schema,
                                existing_coco_ann=anns[ann_id] if ann_id in anns.keys() else None,
                                category_dict=self.category_dict,
                            )
                elif group == SchemaGroup.ANNOTATION:
                    for schema in schema_data:
                        if isinstance(schema, BBox | CompressedRLE):
                            ann_id = f"{schema.view_ref.id}_{schema.entity_ref.id}"
                            anns[ann_id] = coco_annotation(
                                ann=schema,
                                existing_coco_ann=anns[ann_id] if ann_id in anns.keys() else None,
                            )
        export_data["annotations"] = list(anns.values())
        return export_data

    def save_data(self, export_data: dict[str, Any], split: str, file_name: str, file_num: int) -> None:
        """Save data to the specified directory.

        The saved directory has the following structure:
            export_dir/{split}_{file_name}_0.json
                      /...
                      /{split}_{file_name}_{file_num}.json
                      /...
                      /{split}_{file_name}_n.json


        Args:
            export_data: The dictionary containing the data to be saved.
            split: The split of the dataset item being saved.
            file_name: The name of the file to save the data in.
            file_num: The number of the file to save the data in.
        """
        json_path = self.export_dir / f"{split}_{file_name}_{file_num}.json"
        json_path.write_text(json.dumps(jsonable_encoder(export_data), indent=4), encoding="utf-8")


def coco_image(image: Image, view: str) -> dict[str, Any]:
    """Return image in COCO format.

    Args:
        image: Image
        view: Image view

    Returns:
        Image in COCO format
    """
    coco_img = {
        "id": image.id,
        "view": view,
        "width": image.width,
        "height": image.height,
        "file_name": image.url,
        "license": 0,
        "date_captured": image.created_at,
    }
    if isinstance(image, SequenceFrame):
        coco_img["timestamp"] = image.timestamp
        coco_img["frame_index"] = image.frame_index
    return coco_img


def coco_annotation(
    ann: BBox | CompressedRLE | Entity,
    existing_coco_ann: dict[str, Any] | None = None,
    category_dict: dict[str, int] | None = None,
) -> dict[str, Any]:
    """Return annotation in COCO format.

    Args:
        ann: Annotation
        existing_coco_ann: Existing annotation in COCO format to complete
        category_dict: Category dictonary for name to ID conversion
    Returns:
        Annotation in COCO format
    """
    # Load the existing COCO format or initialize a new one
    coco_ann = (
        existing_coco_ann
        if existing_coco_ann is not None
        else {
            "id": f"{ann.view_ref.id}_{ann.id if isinstance(ann, Entity) else ann.entity_ref.id}",
            "image_id": ann.view_ref.id,
            "category_id": None,
            "category_name": None,
            "segmentation": None,
            "area": None,
            "bbox": None,
            "confidence": None,
            "iscrowd": 0,
            "pixano_entity_id": ann.id if isinstance(ann, Entity) else ann.entity_ref.id,
        }
    )
    # Add the specific elements from the annotation into the COCO format
    if isinstance(ann, Entity):
        category_name = str(ann.category).strip().lower()
        coco_ann["category_name"] = category_name
        coco_ann["category_id"] = (
            category_dict[category_name] if category_dict is not None and category_name in category_dict else None
        )
    else:
        if isinstance(ann, BBox):
            coco_ann["pixano_bbox_id"] = ann.id
            coco_ann["bbox"] = ann.xywh_coords
            coco_ann["confidence"] = ann.confidence
        elif isinstance(ann, CompressedRLE):
            coco_ann["pixano_segmentation_id"] = ann.id
            coco_ann["segmentation"] = ann.to_polygons()
            coco_ann["area"] = ann.area
    return coco_ann



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import BaseModel

from pixano.app.models import AnnotationModel, EntityModel, TableInfo, ViewModel
from pixano.app.routers.inference.utils import get_client_from_settings
from pixano.app.routers.utils import get_dataset
from pixano.app.settings import Settings, get_settings
from pixano.features import BBox, Classification, Entity, Image, is_image
from pixano.inference import image_zero_shot_detection

from .utils import get_model_source


router = APIRouter(prefix="/tasks/zero-shot-detection", tags=["Task", "Zero Shot Detection"])


class ZeroShotOutput(BaseModel):
    """Zero shot output."""

    bbox: AnnotationModel
    classification: AnnotationModel


@router.post(
    "/image",
    response_model=list[ZeroShotOutput],
)
async def call_image_zero_shot_detection(
    dataset_id: Annotated[str, Body(embed=True)],
    image: Annotated[ViewModel, Body(embed=True)],
    entity: Annotated[EntityModel, Body(embed=True)],
    classes: Annotated[list[str] | str, Body(embed=True)],
    model: Annotated[str, Body(embed=True)],
    box_table_name: Annotated[str, Body(embed=True)],
    class_table_name: Annotated[str, Body(embed=True)],
    settings: Annotated[Settings, Depends(get_settings)],
    box_threshold: Annotated[float, Body(embed=True)] = 0.3,
    text_threshold: Annotated[float, Body(embed=True)] = 0.2,
) -> list[ZeroShotOutput]:
    """Perform zero shot detection on an image.

    Args:
        dataset_id: The ID of the dataset to use.
        image: The image to use for detection.
        entity: The entity to use for detection.
        classes: Labels to detect.
        model: The name of the model to use.
        box_table_name: The name of the table to use for boxes in dataset.
        class_table_name: The name of the table to use for classifications in dataset.
        settings: App settings.
        box_threshold: Box threshold for detection in the image.
        text_threshold: Text threshold for detection in the image.

    Returns:
        The predicted bboxes and classifications.
    """
    dataset = get_dataset(dataset_id=dataset_id, dir=settings.library_dir, media_dir=settings.media_dir)
    client = get_client_from_settings(settings=settings)

    if not is_image(dataset.schema.schemas[image.table_info.name]):
        raise HTTPException(status_code=400, detail="Image must be an image.")

    image_row: Image = image.to_row(dataset)
    entity_row: Entity = entity.to_row(dataset)
    source = get_model_source(dataset=dataset, model=model)

    try:
        bboxes_and_classifications: list[tuple[BBox, Classification]] = await image_zero_shot_detection(
            client=client,
            source=source,
            media_dir=settings.media_dir,
            image=image_row,
            entity=entity_row,
            classes=classes,
            box_threshold=box_threshold,
            text_threshold=text_threshold,
        )
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e

    output: list[ZeroShotOutput] = []
    for bbox, classification in bboxes_and_classifications:
        bbox_model = AnnotationModel.from_row(
            row=bbox, table_info=TableInfo(name=box_table_name, group="annotations", base_schema="BBox")
        )
        classification_model = AnnotationModel.from_row(
            row=classification,
            table_info=TableInfo(name=class_table_name, group="annotations", base_schema="Classification"),
        )
        output.append(ZeroShotOutput(bbox=bbox_model, classification=classification_model))
    return output



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing_extensions import TYPE_CHECKING

from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..base_schema import BaseSchema
from ..registry import _register_schema_internal


if TYPE_CHECKING:
    from pixano.features.schemas import Item


@_register_schema_internal
class View(BaseSchema):
    """View base class.

    Views are used to define a view in a dataset such as an image, a point cloud, a text. It can refer to an item and
    a parent view.

    Attributes:
        item_ref: Reference to the view's item.
        parent_ref: Reference to the view's parent view.
    """

    item_ref: ItemRef = ItemRef.none()
    parent_ref: ViewRef = ViewRef.none()

    @property
    def item(self) -> "Item":
        """Get the view's item."""
        return self.resolve_ref(self.item_ref)

    @property
    def parent(self) -> "View":
        """Get the view's parent view."""
        return self.resolve_ref(self.parent_ref)


def is_view(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `View` or subclass of `View`."""
    return issubclass_strict(cls, View, strict)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.features import (
    BBox,
    CompressedRLE,
    Conversation,
    Entity,
    Image,
    KeyPoints,
    Message,
    SequenceFrame,
    Track,
    Tracklet,
)

from ..dataset_schema import DatasetItem


class DefaultVQADatasetItem(DatasetItem):
    """Default VQA DatasetItem Schema."""

    image: Image
    conversations: list[Conversation]
    messages: list[Message]
    objects: list[Entity]
    bboxes: list[BBox]
    masks: list[CompressedRLE]
    keypoints: list[KeyPoints]


class DefaultVideoDatasetItem(DatasetItem):
    """Default Video DatasetItem Schema."""

    image: list[SequenceFrame]
    tracks: list[Track]
    tracklets: list[Tracklet]
    bboxes: list[BBox]
    keypoints: list[KeyPoints]


class DefaultImageDatasetItem(DatasetItem):
    """Default Image DatasetItem Schema."""

    image: Image
    objects: list[Entity]
    bboxes: list[BBox]
    masks: list[CompressedRLE]
    keypoints: list[KeyPoints]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from s3path import S3Path

from pixano.app.routers.annotations import router as annotations_router
from pixano.app.routers.browser import router as browser_router
from pixano.app.routers.dataset_items import router as dataset_items_router
from pixano.app.routers.datasets import router as datasets_router
from pixano.app.routers.embeddings import router as embeddings_router
from pixano.app.routers.entities import router as entities_router
from pixano.app.routers.inference import router as inference_router
from pixano.app.routers.items import router as items_router
from pixano.app.routers.items_info import router as items_info_router
from pixano.app.routers.models import router as models_router
from pixano.app.routers.sources import router as sources_router
from pixano.app.routers.views import router as views_router
from pixano.app.settings import Settings


def create_app(settings: Settings = Settings()) -> FastAPI:
    """Create and configure the Pixano app.

    Args:
        settings: App settings.

    Returns:
        The Pixano app.
    """
    # Create app
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=False,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Mount folders
    if isinstance(settings.media_dir, S3Path):
        # If S3, mount models folder
        # Check if folder exists
        if settings.models_dir is None:
            raise FileNotFoundError("Local model directory not provided")
        elif not settings.models_dir.exists():
            raise FileNotFoundError(f"Local model directory '{settings.models_dir.absolute()}' not found")
        # Mount
        app.mount(
            "/models",
            StaticFiles(directory=settings.models_dir),
            name="models",
        )
    else:
        # If local, mount media folder and models folder
        # Check if folder exists
        if not settings.media_dir.exists():
            raise FileNotFoundError(f"Media directory '{settings.media_dir.absolute()}' not found")
        if settings.models_dir is None:
            raise FileNotFoundError("Model directory not provided")
        # Create models folder in case it doesn't exist yet
        if not settings.models_dir.exists():
            settings.models_dir.mkdir(exist_ok=True)
        # Mount
        app.mount(
            "/media",
            StaticFiles(directory=settings.media_dir),
            name="media",
        )
        app.mount(
            "/app_models",
            StaticFiles(directory=settings.models_dir),
            name="models",
        )

    # Include routers
    app.include_router(annotations_router)
    app.include_router(dataset_items_router)
    app.include_router(datasets_router)
    app.include_router(browser_router)
    app.include_router(embeddings_router)
    app.include_router(entities_router)
    app.include_router(items_router)
    app.include_router(items_info_router)
    app.include_router(sources_router)
    app.include_router(views_router)
    app.include_router(models_router)
    app.include_router(inference_router)

    return app



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from functools import lru_cache
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import boto3
from pixano_inference.client import PixanoInferenceClient
from pydantic import field_validator, model_validator
from pydantic_settings import BaseSettings
from s3path import S3Path, register_configuration_parameter
from typing_extensions import Self


class Settings(BaseSettings):
    """Pixano app settings.

    Attributes:
        library_dir: Local or S3 path to dataset library. If not provided, it is set to `'./library/'`.
        media_dir: Local or S3 path to media library. If not provided, it is set to `'./media/'`.
        models_dir: Models directory as Path. Must be provided if library_dir is an S3 path else it is set to
            `'library_dir/models'`
        aws_endpoint: S3 endpoint URL, use 'AWS' if not provided.
            Used if library_dir is an S3 path.
        aws_region: S3 region name, not always required for private storages.
            Used if library_dir is an S3 path.
        aws_access_key: S3 AWS access key. Used if library_dir is an S3 path.
        aws_secret_key: S3 AWS secret key. Used if library_dir is an S3 path.
    """

    library_dir: Path | S3Path = Path.cwd() / "library"
    media_dir: Path | S3Path = Path.cwd() / "media"
    models_dir: Path | None = None
    aws_endpoint: str | None = None
    aws_region: str | None = None
    aws_access_key: str | None = None
    aws_secret_key: str | None = None
    pixano_inference_client: PixanoInferenceClient | None = None

    @field_validator("library_dir", mode="before")
    @classmethod
    def _validate_before_library_dir(cls, library_dir: Any):
        if isinstance(library_dir, str):
            if urlparse(library_dir).scheme == "s3":
                library_dir = S3Path.from_uri(library_dir)
                return library_dir
            return Path(library_dir)
        return library_dir

    @field_validator("media_dir", mode="before")
    @classmethod
    def _validate_before_media_dir(cls, media_dir: Any):
        if isinstance(media_dir, str):
            if urlparse(media_dir).scheme == "s3":
                media_dir = S3Path.from_uri(media_dir)
                return media_dir
            return Path(media_dir)
        return media_dir

    @field_validator("models_dir", mode="before")
    @classmethod
    def _validate_before_model_dir(cls, models_dir: Any):
        if isinstance(models_dir, str):
            return Path(models_dir)
        return models_dir

    @model_validator(mode="after")
    def _validate_after_model(self) -> Self:
        # Setup library and media directories
        for attr in ["library_dir", "media_dir"]:
            if isinstance(getattr(self, attr), S3Path):
                try:
                    register_configuration_parameter(
                        getattr(self, attr),
                        resource=boto3.resource(
                            "s3",
                            endpoint_url=self.aws_endpoint,
                            region_name=self.aws_region,
                            aws_access_key_id=self.aws_access_key,
                            aws_secret_access_key=self.aws_secret_key,
                        ),
                    )
                except Exception as e:
                    raise ValueError(
                        f"ERROR: Could not register S3 {attr} library.\n"
                        "You have to set the following environment variables:\n"
                        "- AWS_ENDPOINT: S3 endpoint URL, use 'AWS' if not provided\n"
                        "- AWS_REGION: S3 region name, not always required "
                        "for private storages\n"
                        "- AWS_ACCESS_KEY: S3 AWS access key\n"
                        "- AWS_SECRET_KEY: S3 AWS secret key"
                    ) from e

        # Check if local model directory is provided
        if isinstance(self.library_dir, S3Path) and self.models_dir is None:
            raise AttributeError(
                "When using S3 storage for the library, runtime models (.onnx files) must be "
                "stored locally and their directory must be provided with 'model_dir'."
            )
        elif self.models_dir is None:
            self.models_dir = self.library_dir / "models"
        return self


@lru_cache
def get_settings() -> Settings:
    """Get app settings.

    Returns:
        App settings.
    """
    return Settings()



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import BaseModel

from pixano.app.models import AnnotationModel, EntityModel, TableInfo, ViewModel
from pixano.app.routers.inference.utils import get_client_from_settings
from pixano.app.routers.utils import get_dataset
from pixano.app.settings import Settings, get_settings
from pixano.features import BBox, CompressedRLE, Entity, Image, NDArrayFloat, is_image
from pixano.inference import image_mask_generation

from .utils import get_model_source


router = APIRouter(prefix="/tasks/mask-generation", tags=["Task", "Mask Generation"])

# keep track of the last image used, to avoid recomputing embeddings if not necessary
store_last_image_id: str = ""


class ImageMaskGenerationOutput(BaseModel):
    """Image mask generation output."""

    mask: AnnotationModel


@router.post(
    "/image",
    response_model=ImageMaskGenerationOutput,
)
async def call_image_mask_generation(
    dataset_id: Annotated[str, Body(embed=True)],
    image: Annotated[ViewModel, Body(embed=True)],
    model: Annotated[str, Body(embed=True)],
    mask_table_name: Annotated[str, Body(embed=True)],
    settings: Annotated[Settings, Depends(get_settings)],
    entity: Annotated[EntityModel | None, Body(embed=True)] = None,
    bbox: Annotated[BBox | None, Body(embed=True)] = None,
    points: Annotated[list[list[int]] | None, Body(embed=True)] = None,
    labels: Annotated[list[int] | None, Body(embed=True)] = None,
) -> ImageMaskGenerationOutput:
    """Perform image mask generation on an image.

    Args:
        dataset_id: The ID of the dataset to use.
        image: The image to use for detection.
        entity: The entity to use for detection.
        model: The name of the model to use.
        mask_table_name: The name of the table to use for masks in dataset.
        settings: App settings.
        bbox: Input bounding box or None.
        points: Input points or None.
        labels: Labels for input points, or None.

    Returns:
        The generated mask.
    """
    global store_last_image_id

    dataset = get_dataset(dataset_id=dataset_id, dir=settings.library_dir, media_dir=settings.media_dir)
    client = get_client_from_settings(settings=settings)

    if not is_image(dataset.schema.schemas[image.table_info.name]):
        raise HTTPException(status_code=400, detail="Image must be an image.")

    image_row: Image = image.to_row(dataset)
    entity_row: Entity | None = entity.to_row(dataset) if entity else None
    source = get_model_source(dataset=dataset, model=model)

    try:
        mask_output: tuple[
            CompressedRLE, float, NDArrayFloat | None, list[NDArrayFloat] | None
        ] = await image_mask_generation(
            client=client,
            source=source,
            media_dir=settings.media_dir,
            image=image_row,
            entity=entity_row,
            image_embedding=None,
            high_resolution_features=None,
            reset_predictor=store_last_image_id != image_row.id,
            bbox=bbox,
            points=points,
            labels=labels,
        )
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e

    # store image_id to check against for next generation
    store_last_image_id = image_row.id

    output: ImageMaskGenerationOutput = ImageMaskGenerationOutput(
        mask=AnnotationModel.from_row(
            row=mask_output[0],  # right now we don't use other ouputs (score, ...)
            table_info=TableInfo(name=mask_table_name, group="annotations", base_schema="CompressedRLE"),
        )
    )
    return output



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .base_type import BaseType, is_base_type
from .nd_array_float import NDArrayFloat, create_ndarray_float, is_ndarray_float
from .schema_reference import (
    AnnotationRef,
    EmbeddingRef,
    EntityRef,
    ItemRef,
    SchemaRef,
    SourceRef,
    ViewRef,
    create_annotation_ref,
    create_embedding_ref,
    create_entity_ref,
    create_item_ref,
    create_schema_ref,
    create_source_ref,
    create_view_ref,
    is_annotation_ref,
    is_embedding_ref,
    is_entity_ref,
    is_item_ref,
    is_schema_ref,
    is_source_ref,
    is_view_ref,
)


__all__ = [
    "AnnotationRef",
    "BaseType",
    "EmbeddingRef",
    "EntityRef",
    "ItemRef",
    "NDArrayFloat",
    "ViewRef",
    "SchemaRef",
    "SourceRef",
    "create_annotation_ref",
    "create_embedding_ref",
    "create_entity_ref",
    "create_item_ref",
    "create_ndarray_float",
    "create_schema_ref",
    "create_source_ref",
    "create_view_ref",
    "is_annotation_ref",
    "is_base_type",
    "is_embedding_ref",
    "is_entity_ref",
    "is_item_ref",
    "is_ndarray_float",
    "is_schema_ref",
    "is_source_ref",
    "is_view_ref",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .embedding import (
    Embedding,
    ViewEmbedding,
    create_view_embedding_function,
    is_embedding,
    is_view_embedding,
)


__all__ = [
    "Embedding",
    "ViewEmbedding",
    "is_embedding",
    "is_view_embedding",
    "create_view_embedding_function",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from enum import Enum

from .dataset_items import DefaultImageDatasetItem, DefaultVideoDatasetItem, DefaultVQADatasetItem


class WorkspaceType(Enum):
    """Workspace type.

    A workspace is a specific environment where the dataset is used. It defines what

    Attributes:
        IMAGE: Image workspace.
        VIDEO: Video workspace.
        IMAGE_VQA: Image VQA workspace.
        IMAGE_TEXT_ENTITY_LINKING: Image text entity linking workspace.
        UNDEFINED: Undefined workspace.
    """

    IMAGE = "image"
    VIDEO = "video"
    IMAGE_VQA = "image_vqa"
    IMAGE_TEXT_ENTITY_LINKING = "image_text_entity_linking"
    UNDEFINED = "undefined"


__all__ = [
    "WorkspaceType",
    "DefaultImageDatasetItem",
    "DefaultVideoDatasetItem",
    "DefaultVQADatasetItem",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .python import estimate_folder_size, get_super_type_from_dict, natural_key, unique_list
from .validation import issubclass_strict, validate_and_init_create_at_and_update_at


__all__ = [
    "estimate_folder_size",
    "unique_list",
    "get_super_type_from_dict",
    "issubclass_strict",
    "natural_key",
    "validate_and_init_create_at_and_update_at",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, Query

from pixano.app.models.items import ItemModel
from pixano.app.settings import Settings, get_settings
from pixano.features.schemas.schema_group import SchemaGroup

from .utils import (
    create_row_handler,
    create_rows_handler,
    delete_row_handler,
    delete_rows_handler,
    get_row_handler,
    get_rows_handler,
    update_row_handler,
    update_rows_handler,
)


router = APIRouter(prefix="/items", tags=["Items"])


@router.get("/{dataset_id}/", response_model=list[ItemModel])
async def get_items(
    dataset_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
    where: str | None = None,
) -> list[ItemModel]:
    """Get sources from the `'item'` table of a dataset.

    They can be filtered by IDs, a where clause or paginated.

    Args:
        dataset_id: Dataset ID containing the table.
        settings: App settings.
        ids: IDs of the sources.
        limit: Limit number of sources.
        skip: Skip number of sources.
        where: Where clause.

    Returns:
        List of sources.
    """
    return await get_rows_handler(
        dataset_id=dataset_id,
        group=SchemaGroup.ITEM,
        table=SchemaGroup.ITEM.value,
        settings=settings,
        where=where,
        ids=ids,
        item_ids=None,
        limit=limit,
        skip=skip,
    )


@router.get("/{dataset_id}/{id}", response_model=ItemModel)
async def get_item(dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]) -> ItemModel:
    """Get an item from the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the item.
        settings: App settings.

    Returns:
        The item.
    """
    return await get_row_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, id, settings)


@router.post("/{dataset_id}/", response_model=list[ItemModel])
async def create_items(
    dataset_id: str,
    items: list[ItemModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[ItemModel]:
    """Add items in the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        items: Items to add.
        settings: App settings.

    Returns:
        List of items added.
    """
    return await create_rows_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, items, settings)


@router.post("/{dataset_id}/{id}", response_model=ItemModel)
async def create_item(
    dataset_id: str,
    id: str,
    item: ItemModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> ItemModel:
    """Add an item in the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the item.
        item: Item to add.
        settings: App settings.

    Returns:
        The item added.
    """
    return await create_row_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, id, item, settings)


@router.put("/{dataset_id}/{id}", response_model=ItemModel)
async def update_item(
    dataset_id: str,
    id: str,
    item: ItemModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> ItemModel:
    """Update an item in the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the item.
        item: Item to update.
        settings: App settings.

    Returns:
        The item updated.
    """
    return await update_row_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, id, item, settings)


@router.put("/{dataset_id}/", response_model=list[ItemModel])
async def update_items(
    dataset_id: str,
    items: list[ItemModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[ItemModel]:
    """Update items in the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        items: Items to update.
        settings: App settings.

    Returns:
        List of items updated.
    """
    return await update_rows_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, items, settings)


@router.delete("/{dataset_id}/{id}")
async def delete_item(dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]) -> None:
    """Delete an item from the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the item to delete.
        settings: App settings.
    """
    return await delete_row_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, id, settings)


@router.delete("/{dataset_id}/")
async def delete_items(
    dataset_id: str,
    ids: Annotated[list[str], Query()],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete items from the `'item'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        ids: IDs of the items to delete.
        settings: App settings.
    """
    return await delete_rows_handler(dataset_id, SchemaGroup.ITEM, SchemaGroup.ITEM.value, ids, settings)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Literal

from pydantic import model_validator

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@_register_schema_internal
class KeyPoints(Annotation):
    """A set of keypoints.

    Attributes:
        template_id: Id of the keypoint template.
        coords: List of 2D coordinates of the keypoints.
        states: Status for each keypoint. ("visible", "invisible", "hidden").
    """

    template_id: str
    coords: list[float]
    states: list[str]  # replace by features: list[dict] ?

    @model_validator(mode="after")
    def _validate_fields(self):
        if len(self.coords) % 2 != 0:
            raise ValueError("There must be an even number of coords")
        elif any(coord < 0 for coord in self.coords):
            raise ValueError("Coordinates must be positive")
        elif len(self.states) != len(self.coords) // 2:
            raise ValueError("There must be the same number of states than points")
        elif any(state not in ["visible", "invisible", "hidden"] for state in self.states):
            raise ValueError("States must be 'visible', 'invisible' or 'hidden'")
        return self

    @classmethod
    def none(cls) -> "KeyPoints":
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" KeyPoints.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            template_id="",
            coords=[0, 0],
            states=["invisible"],
        )

    def map_back2front_vertices(self) -> list:
        """Utility function to map back format for KeyPoint to front vertices format.

        Raises:
            ValueError: If keypoints is ill-formed.

        Returns:
            keypoint list for vertices front format.
        """
        # Check coords are even
        if len(self.coords) % 2 != 0:
            raise ValueError("There must be an even number of coords")

        result = []
        if self.states is not None:
            num_points = len(self.coords) // 2
            if len(self.states) != num_points:
                raise ValueError("There must be the same number of states than points")

            result = [
                {"x": x, "y": y, "features": {"state": state}}
                for (x, y), state in zip(zip(self.coords[0::2], self.coords[1::2]), self.states)
            ]
        else:
            result = [{"x": x, "y": y} for x, y in zip(self.coords[0::2], self.coords[1::2])]
        return result


@_register_schema_internal
class KeyPoints3D(Annotation):
    """A set of 3D keypoints.

    Attributes:
        template_id: id of keypoint template.
        coords: List of 3D coordinates of the keypoints.
        states: Status for each keypoint.
    """

    template_id: str
    coords: list[float]
    states: list[str]  # replace by features: list[dict] ?

    @model_validator(mode="after")
    def _validate_fields(self):
        if len(self.coords) % 3 != 0:
            raise ValueError("There must be a multiple of 3 coords")
        elif any(coord < 0 for coord in self.coords):
            raise ValueError("Coordinates must be positive")
        if len(self.states) != len(self.coords) // 3:
            raise ValueError("There must be the same number of states than points")
        elif any(state not in ["visible", "invisible", "hidden"] for state in self.states):
            raise ValueError("States must be 'visible', 'invisible' or 'hidden'")
        return self

    @classmethod
    def none(cls) -> "KeyPoints3D":
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" KeyPoints3D.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            template_id="",
            coords=[0, 0, 0],
            states=["visible"],
        )

    def map_back2front_vertices(self) -> list:
        """Utility function to map back format for KeyPoint3D to front vertices format.

        Warn:
            Not implemented for 3D keypoints.
        """
        raise NotImplementedError("Not implemented for 3D keypoints.")


def is_keypoints(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `KeyPoints` or subclass of `KeyPoints`."""
    return issubclass_strict(cls, KeyPoints, strict)


def is_keypoints3d(cls: type, strict: bool = False) -> bool:
    """Check if a class is `Keypoints3D` or a subclass of `Keypoints3D`."""
    return issubclass_strict(cls, KeyPoints3D, strict)


def create_keypoints(
    template_id: str,
    coords: list[float],
    states: list[str],
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> KeyPoints:
    """Create a `KeyPoints` instance.

    Args:
        template_id: id of keypoint template.
        coords: List of 2D coordinates of the keypoints.
        states: Status for each keypoint. ("visible", "invisible", "hidden").
        id: Keypoints ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `KeyPoints` instance.
    """
    return KeyPoints(
        template_id=template_id,
        coords=coords,
        states=states,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )


def create_keypoints3d(
    template_id: str,
    coords: list[float],
    states: list[Literal["visible", "invisble", "hidden"]],
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> KeyPoints3D:
    """Create a `KeyPoints3D` instance.

    Args:
        template_id: The id of the keypoint template.
        coords: The 3D coordinates of the keypoints.
        states: The visibility status for each keypoint.
        id: Keypoints3D ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `KeyPoints3D` instance.
    """
    return KeyPoints3D(
        template_id=template_id,
        coords=coords,
        states=states,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .annotations import AnnotationModel
from .base_schema import BaseSchemaModel
from .dataset_info import DatasetInfoModel
from .dataset_items import DatasetItemModel
from .datasets import DatasetBrowser, DatasetModel, PaginationColumn, PaginationInfo, TableData
from .embeddings import EmbeddingModel
from .entities import EntityModel
from .item_info import ItemInfoModel
from .items import ItemModel
from .sources import SourceModel
from .table_info import TableInfo
from .views import ViewModel


__all__ = [
    "AnnotationModel",
    "BaseSchemaModel",
    "PaginationColumn",
    "DatasetBrowser",
    "DatasetInfoModel",
    "DatasetItemModel",
    "DatasetModel",
    "EmbeddingModel",
    "EntityModel",
    "ItemInfoModel",
    "ItemModel",
    "PaginationInfo",
    "SourceModel",
    "TableData",
    "TableInfo",
    "ViewModel",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..registry import _register_schema_internal
from .view import View


@_register_schema_internal
class PointCloud(View):
    """Point Cloud view.

    Attributes:
        url: The point cloud URL.
    """

    url: str


def is_point_cloud(cls: type, strict: bool = False) -> bool:
    """Check if the given class is a subclass of `PointCloud`."""
    return issubclass_strict(cls, PointCloud, strict)


def create_point_cloud(
    url: str, id: str = "", item_ref: ItemRef = ItemRef.none(), parent_ref: ViewRef = ViewRef.none()
) -> PointCloud:
    """Create a `PointCloud` instance.

    Args:
        url: The point cloud URL.
        id: Point cloud ID.
        item_ref: Item reference.
        parent_ref: Parent view reference.

    Returns:
        The created `PointCloud` instance.
    """
    return PointCloud(url=url, id=id, item_ref=item_ref, parent_ref=parent_ref)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import re
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator, Mapping

import pyarrow.json as pa_json
import shortuuid

from pixano.datasets.dataset_info import DatasetInfo
from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.utils import mosaic
from pixano.datasets.workspaces import WorkspaceType
from pixano.features import (
    BaseSchema,
    Entity,
    Item,
    Message,
    View,
    create_bbox,
    create_conversation,
    is_annotation,
    is_bbox,
    is_conversation,
    is_entity,
    is_view,
)
from pixano.features.schemas.annotations import Annotation
from pixano.features.schemas.registry import _PIXANO_SCHEMA_REGISTRY
from pixano.features.schemas.source import SourceKind
from pixano.features.types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from pixano.features.utils.creators import create_instance_of_schema

from ..dataset_builder import DatasetBuilder


class FolderBaseBuilder(DatasetBuilder):
    """This is a class for building datasets based on a folder structure.

    The folder structure should be as follows:
        - source_dir/{split}/{item}.{ext}
        - source_dir/{split}/metadata.jsonl

    The metadata file should be a jsonl file with the following format:
    ```json
        [
            {
                "item": "item1",
                "metadata1": "value1",
                "metadata2": "value2",
                ...
                "entities": {
                    "attr1": [val1, val2, ...],
                    "attr2": [val1, val2, ...],
                    ...
                }
            },
            {
                "item": "item2",
                "metadata1": "value1",
                "metadata2": "value2",
                ...
                "entities": {
                    "attr1": [val1, val2, ...],
                    "attr2": [val1, val2, ...],
                    ...
                }
            },
            ...
        ]
    ```

    Note:
        Only one view is supported in folder builders. If you give a list of images, it will be put in a mosaic.

    Attributes:
        source_dir: The source directory for the dataset.
        view_name: The name of the view schema.
        view_schema: The schema of the view.
        entity_name: The name of the entities schema.
        entity_schema: The schema of the entities.
        METADATA_FILENAME: The metadata filename.
        EXTENSIONS: The list of supported extensions.
        WORKSPACE_TYPE: The workspace type of the dataset. Subclass should override this attribute if workspace is
            known.
    """

    METADATA_FILENAME: str = "metadata.jsonl"
    EXTENSIONS: list[str]
    WORKSPACE_TYPE = WorkspaceType.UNDEFINED
    DEFAULT_SCHEMA: type[DatasetItem] | None = None

    def __init__(
        self,
        media_dir: Path | str,
        library_dir: Path | str,
        info: DatasetInfo,
        dataset_path: Path | str,
        dataset_item: type[DatasetItem] | None = None,
        use_image_name_as_id: bool = False,
    ) -> None:
        """Initialize the `FolderBaseBuilder`.

        Args:
            media_dir: The global media directory.
            library_dir: The global directory for Pixano datasets library.
            dataset_item: The dataset item schema.
            info: User informations (name, description, ...) for the dataset.
            dataset_path: Path to dataset, relative to media_dir.
            use_image_name_as_id: If True, use image base name as image id.
                                  Images MUST have unique base names.
                                  When no metadata file exists, also use it as item id,
                                  else, use 'item_#'
                                  This allows to reuse image embeddings after dataset overwrite.
        """
        info.workspace = self.WORKSPACE_TYPE
        if self.DEFAULT_SCHEMA is not None and dataset_item is None:
            dataset_item = self.DEFAULT_SCHEMA
        if dataset_item is None:
            raise ValueError("A schema is required.")

        self.use_image_name_as_id = use_image_name_as_id

        self.media_dir = Path(media_dir)
        dataset_path = Path(dataset_path)
        self.source_dir = self.media_dir / dataset_path
        if not self.source_dir.is_dir():
            raise ValueError("A source path (media_dir / dataset_path) is required.")

        target_dir = Path(library_dir) / "_".join(dataset_path.parts)
        super().__init__(target_dir=target_dir, dataset_item=dataset_item, info=info)

        self.views_schema: dict[str, type[View]] = {}
        self.entities_schema: dict[str, type[Entity]] = {}
        self.annotations_schema: dict[str, type[Annotation]] = {}

        for k, s in self.schemas.items():
            if is_view(s):
                self.views_schema.update({k: s})
            elif is_entity(s):
                self.entities_schema.update({k: s})
            elif is_annotation(s):
                self.annotations_schema.update({k: s})
        if not self.views_schema or not self.entities_schema:
            raise ValueError("At least one View and one Entity schema must be defined in the schemas argument.")

        # TODO - allow multiview in base FolderBuilder
        if len(self.views_schema) > 1:
            raise ValueError("Only one view schema is supported in folder based builders.")

    def generate_data(
        self,
    ) -> Iterator[dict[str, BaseSchema | list[BaseSchema]]]:
        """Generate data from the source directory.

        Returns:
            An iterator over the data following the dataset schemas.
        """
        self.source_id = self.add_source("Builder", SourceKind.OTHER)
        for split in self.source_dir.glob("*"):
            if not split.is_dir() or split.name.startswith("."):
                continue

            try:
                dataset_pieces = self._read_metadata(split / self.METADATA_FILENAME)
            except FileNotFoundError:
                dataset_pieces = None

            if dataset_pieces is None:
                for view_file in sorted(split.glob("**/*")):
                    # only consider {split}/**/{item}.{ext} files
                    if not view_file.is_file() or view_file.suffix not in self.EXTENSIONS:
                        continue
                    # create item with default values for custom fields
                    custom_item_metadata = self._build_default_custom_metadata_item()
                    custom_item_metadata["id"] = view_file.stem
                    custom_item_metadata["split"] = split.name
                    item = self._create_item(**custom_item_metadata)
                    # create view
                    view_name_nojsonl, view_schema_nojsonl = list(self.views_schema.items())[0]  # only one view
                    view = self._create_view(item, view_file, view_schema_nojsonl)
                    yield {
                        self.item_schema_name: item,
                        view_name_nojsonl: view,
                    }
                    # if schema contain a Conversation, add one
                    for entity_name, entity_schema_nojsonl in self.entities_schema.items():
                        if entity_schema_nojsonl is not None and is_conversation(entity_schema_nojsonl):
                            default_view_ref = ViewRef(id=view.id, name=view_name_nojsonl)
                            conversation = create_conversation(
                                id=shortuuid.uuid(),
                                kind="vqa",
                                item_ref=ItemRef(id=item.id),
                                view_ref=default_view_ref,
                            )
                            yield {"conversations": conversation}

                continue

            for i, dataset_piece in enumerate(dataset_pieces):
                item_metadata = {}
                for k in dataset_piece.keys():
                    if (
                        k not in self.views_schema
                        and k not in self.entities_schema
                        and k not in self.annotations_schema
                    ):
                        item_metadata.update({k: dataset_piece.get(k, None)})
                for k in item_metadata.keys():
                    dataset_piece.pop(k, None)

                # create item
                if "id" not in item_metadata:
                    item_metadata["id"] = f"item_{split.name}_{i}" if self.use_image_name_as_id else shortuuid.uuid()
                if "split" not in item_metadata:
                    item_metadata["split"] = split.name
                item = self._create_item(**item_metadata)

                # create view
                views_data: list[tuple[str, View]] = []
                for k, v in dataset_piece.items():
                    if k in self.views_schema:
                        view_name = k
                        view_schema = self.views_schema.get(view_name)
                        if view_schema is not None:
                            if isinstance(v, list):
                                if len(v) == 0:
                                    continue
                                if len(v) > 1:
                                    # create a mosaic from item images
                                    mosaic_file = mosaic(self.source_dir, split.name, v, view_name)
                                    view_file = self.source_dir / mosaic_file
                                    if not view_file.is_file():  # no split path in metadata.jsonl
                                        view_file = self.source_dir / split.name / mosaic_file
                                else:
                                    view_file = self.source_dir / Path(v[0])
                                    if not view_file.is_file():  # no split path in metadata.jsonl
                                        view_file = self.source_dir / split.name / Path(v[0])
                                if view_file.is_file() and view_file.suffix in self.EXTENSIONS:
                                    view = self._create_view(item, view_file, view_schema)
                                    views_data.append((view_name, view))
                            else:
                                view_file = self.source_dir / (
                                    Path(v) if split.name == Path(v).parts[0] else split / Path(v)
                                )
                                if view_file.is_file() and view_file.suffix in self.EXTENSIONS:
                                    view = self._create_view(item, view_file, view_schema)
                                    views_data.append((view_name, view))

                all_entities_data: dict[str, list[Entity]] = defaultdict(list)
                all_annotations_data: dict[str, list[Annotation]] = defaultdict(list)
                for k, v in dataset_piece.items():
                    if k in self.entities_schema and v is not None:
                        entity_name = k
                        raw_entities_data = v
                        entity_schema = self.entities_schema.get(entity_name)
                        if entity_schema is not None:
                            if is_conversation(entity_schema):
                                entities_data, annotations_data = self._create_vqa_entities(
                                    item, views_data, entity_name, entity_schema, raw_entities_data
                                )
                            else:  # classic entity
                                entities_data, annotations_data = self._create_objects_entities(
                                    item, views_data, entity_name, entity_schema, raw_entities_data
                                )

                            for name, entities in entities_data.items():
                                all_entities_data[name].extend(entities)

                            for name, annotations in annotations_data.items():
                                all_annotations_data[name].extend(annotations)

                yield {self.item_schema_name: item}
                for view_name, view in views_data:
                    yield {view_name: view}

                if all_entities_data is None:
                    continue

                yield all_entities_data
                yield all_annotations_data

    def _create_item(self, **item_metadata) -> BaseSchema:
        return self.item_schema(**item_metadata)

    def _create_view(self, item: Item, view_file: Path, view_schema: type[View]) -> View:
        if not issubclass(view_schema, View):
            raise ValueError("View schema must be a subclass of View")
        if view_schema not in _PIXANO_SCHEMA_REGISTRY.values():
            raise ValueError(
                f"View schema {view_schema} is not supported. You should implement your own _create_view method."
            )

        view = create_instance_of_schema(
            view_schema,
            id=view_file.stem if self.use_image_name_as_id else shortuuid.uuid(),
            item_ref=ItemRef(id=item.id),
            url=view_file,
            url_relative_path=self.media_dir,
        )
        return view

    def _create_vqa_entities(
        self,
        item: Item,
        views_data: list[tuple[str, View]],
        entity_name: str,
        entity_schema: type[Entity],
        raw_entities_data: list[Any],
    ) -> tuple[dict[str, list[Entity]], dict[str, list[Annotation]]]:
        def update_viewref(content, views_data, view_ref):
            match = re.match(r".*<image (\d)>.*", content)  # TODO image_regex as parameter
            if match is None:
                return view_ref
            for m in match.groups():
                num = int(m) - 1
                if num >= 0 and num < len(views_data):
                    view_name, view = views_data[num]
                    return ViewRef(id=view.id, name=view_name)
            return view_ref

        entities: dict[str, list[Entity]] = defaultdict(list)
        annotations: dict[str, list[Annotation]] = defaultdict(list)

        default_user = "import"
        entities_data: list[list[Any]] = []

        if isinstance(raw_entities_data, list):
            if all(isinstance(data, list) for data in raw_entities_data):
                entities_data = raw_entities_data
            if all(isinstance(data, dict) for data in raw_entities_data):
                entities_data = [raw_entities_data]

        for conv in entities_data:
            view_name, view = views_data[0]  # Only one view
            default_view_ref = ViewRef(id=view.id, name=view_name)

            conversation = entity_schema(
                id=shortuuid.uuid(), kind="vqa", item_ref=ItemRef(id=item.id), view_ref=default_view_ref
            )
            entities[entity_name].append(conversation)

            for i, q_n_a in enumerate(conv):
                question = q_n_a.get("question")
                if not question:
                    raise ValueError(f"No question found for item {item}")

                # Handle question
                if "user" not in question:
                    question["user"] = default_user

                content = question.get("content")
                if not content:
                    raise ValueError(f"No text content found for question in item {item}")

                view_ref = update_viewref(content, views_data, default_view_ref)

                query_msg = Message(
                    id=shortuuid.uuid(),
                    number=i,
                    type="QUESTION",
                    item_ref=ItemRef(id=item.id),
                    entity_ref=EntityRef(id=conversation.id, name=entity_name),
                    source_ref=SourceRef(id=self.source_id),
                    view_ref=view_ref,
                    timestamp=datetime.now(),
                    **question,
                )
                annotations["messages"].append(query_msg)

                # Handle all responses

                responses = []
                response = q_n_a.get("response", None)
                if isinstance(response, dict):
                    responses = [response]
                if not responses:
                    responses = q_n_a.get("responses", [])
                    if not isinstance(responses, list):
                        responses = [responses]

                for response in responses:
                    if "user" not in response:
                        response["user"] = default_user
                    content = response.get("content", "")
                    explanation = response.get("explanation", "")
                    if query_msg.question_type != "OPEN":
                        response["content"] = "[[{}]] {}".format(content, explanation)

                    response_msg = Message(
                        id=shortuuid.uuid(),
                        number=i,
                        type="ANSWER",
                        question_type=query_msg.question_type,
                        item_ref=ItemRef(id=item.id),
                        entity_ref=EntityRef(id=conversation.id, name=entity_name),
                        source_ref=SourceRef(id=self.source_id),
                        timestamp=datetime.now(),
                        **response,
                    )
                    annotations["messages"].append(response_msg)
        return entities, annotations

    def _create_objects_entities(
        self,
        item: Item,
        views_data: list[tuple[str, View]],
        entity_name: str,
        entity_schema: type[Entity],
        entities_data: dict[str, Any],
    ) -> tuple[dict[str, list[Entity]], dict[str, list[Annotation]]]:
        entities: dict[str, list[Entity]] = defaultdict(list)
        annotations: dict[str, list[Annotation]] = defaultdict(list)

        # only one view
        view_name, view = views_data[0]

        entities_attrs = list(entities_data.keys())
        for attr in entities_attrs:
            if attr not in entity_schema.model_fields.keys() and attr not in self.schemas.keys():
                raise ValueError(f"Attribute {attr} not found in entity schema and is not a schema annotation.")

        # check if all list of entities data have same length
        nums_entities = {len(v) if isinstance(v, list) else 1 for v in entities_data.values() if v is not None}
        if len(nums_entities) > 1:
            raise ValueError("All list of entities data must have same length")
        elif len(nums_entities) == 0:
            return {}, {}

        num_entities = nums_entities.pop()
        entities_data = {k: v if isinstance(v, list) else [v] for k, v in entities_data.items() if v is not None}
        entities_attrs = list(entities_data.keys())

        for i in range(num_entities):
            entity = {}
            entity_annotations: dict[str, Any] = {}
            entity_id = shortuuid.uuid()
            for attr in entities_attrs:
                if entities_data[attr] is None:
                    continue

                # check if attribute is an annotation schema
                is_attr_schema = attr in self.schemas
                if is_attr_schema and not issubclass(self.schemas[attr], Annotation):
                    raise ValueError(f"Attribute {attr} must be a subclass of Annotation")

                # create annotation if attribute is an annotation schema
                if is_attr_schema:
                    if attr not in entity_annotations:
                        entity_annotations[attr] = []

                    schema = self.schemas[attr]
                    if isinstance(entities_data[attr][i], Mapping):
                        annotation = create_instance_of_schema(
                            schema,
                            id=shortuuid.uuid(),
                            item_ref=ItemRef(id=item.id),
                            view_ref=ViewRef(id=view.id, name=view_name),
                            entity_ref=EntityRef(id=entity_id, name=entity_name),
                            source_ref=SourceRef(id=self.source_id),
                            **entities_data[attr][i],
                        )
                    else:
                        # TODO check jsonl format for mask & keypoints
                        if is_bbox(schema, True):
                            annotation = create_bbox(
                                id=shortuuid.uuid(),
                                item_ref=ItemRef(id=item.id),
                                view_ref=ViewRef(id=view.id, name=view_name),
                                entity_ref=EntityRef(id=entity_id, name=entity_name),
                                source_ref=SourceRef(id=self.source_id),
                                coords=entities_data[attr][i],
                                format="xywh",
                                is_normalized=all(0 <= x <= 1 for x in entities_data[attr][i]),
                                confidence=1.0,
                            )
                        else:
                            raise ValueError(f"Schema {schema} not supported for infered entity creation.")
                    entity_annotations[attr].append(annotation)
                else:
                    entity[attr] = entities_data[attr][i]
            entities[entity_name].append(
                entity_schema(
                    id=entity_id,
                    item_ref=ItemRef(id=item.id),
                    view_ref=ViewRef(id=view.id, name=view_name),
                    **entity,
                )
            )

            for key, entity_annotation in entity_annotations.items():
                annotations[key].extend(entity_annotation)
        return entities, annotations

    def _read_metadata(self, metadata_file: Path) -> list[dict]:
        if not metadata_file.exists():
            raise FileNotFoundError(f"Metadata file {metadata_file} not found")
        return pa_json.read_json(metadata_file).to_pylist()

    def _build_default_custom_metadata_item(self) -> dict[str, Any]:
        custom_item_metadata: dict[str, Any] = {}
        custom_fields = list(set(self.item_schema.field_names()) - set(Item.field_names()))
        for field in custom_fields:
            field_type = self.item_schema.__annotations__[field]
            custom_item_metadata[field] = field_type()
        return custom_item_metadata



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json

from pydantic import ConfigDict, field_validator
from typing_extensions import Self, TypeVar

from pixano.datasets import Dataset
from pixano.features import Annotation, SchemaGroup, is_annotation

from .base_schema import BaseSchemaModel
from .table_info import TableInfo


T = TypeVar("T", bound=Annotation)


class AnnotationModel(BaseSchemaModel[Annotation]):
    """Model for the [Annotation][pixano.features.Annotation] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "bbox_orange_cat_n1",
                    "table_info": {"group": "annotations", "name": "bboxes", "base_schema": "BBox"},
                    "data": {
                        "item_ref": {"name": "item", "id": "1"},
                        "view_ref": {"name": "image", "id": "orange_cats"},
                        "entity_ref": {
                            "name": "cats",
                            "id": "cat_n1",
                        },
                        "coords": [0, 0, 2, 2],
                        "format": "xywh",
                        "is_normalized": False,
                        "confidence": 0.8,
                        "inference_metadata": {},
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.ANNOTATION.value:
            raise ValueError(f"Table info group must be {SchemaGroup.ANNOTATION.value}.")
        return value

    def to_row(self, dataset: Dataset) -> Annotation:
        """Create an [Annotation][pixano.features.Annotation] from the model."""
        if not is_annotation(dataset.schema.schemas[self.table_info.name]):
            raise ValueError(f"Schema type must be a subclass of {Annotation.__name__}.")
        row = super().to_row(dataset)
        row.inference_metadata = json.dumps(self.data["inference_metadata"])
        return row

    @classmethod
    def from_row(cls, row: Annotation, table_info: TableInfo) -> Self:
        """Create an AnnotationModel from an [Annotation][pixano.features.Annotation].

        Args:
            row: The row to create the model from.
            table_info: The table info of the row.

        Returns:
            The created model.
        """
        annotation_model = BaseSchemaModel.from_row(row, table_info)
        annotation_model.data["inference_metadata"] = json.loads(row.inference_metadata)
        return cls.model_construct(**annotation_model.__dict__)  # Avoid validation and casting



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import base64
import io
from io import BytesIO
from itertools import groupby

import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from pycocotools import mask as mask_api


def image_to_binary(image: Image.Image, im_format: str = "PNG") -> bytes:
    """Encode an image from Pillow to binary.

    Args:
        image: Pillow image.
        im_format: Image file extension.

    Returns:
        Image as binary.
    """
    with BytesIO() as output_bytes:
        image.save(output_bytes, im_format)
        im_bytes = output_bytes.getvalue()

    return im_bytes


def binary_to_url(im_bytes: bytes) -> str:
    """Encode image from binary to base 64 URL.

    Args:
        im_bytes: Image as binary.

    Returns:
        Image base 64 URL.
    """
    encoded = base64.b64encode(im_bytes).decode("utf-8")
    return f"data:image;base64,{encoded}"


def depth_file_to_binary(depth_path: str) -> bytes:
    """Encode depth file to RGB image in binary.

    Args:
        depth_path: Depth file path.

    Returns:
        Depth file as RGB image in binary.
    """
    depth = cv2.imread(depth_path, cv2.IMREAD_ANYDEPTH).astype(np.float32)
    depth = depth_array_to_gray(depth)
    depth_rgb = Image.fromarray(depth)

    return image_to_binary(depth_rgb)


def depth_array_to_gray(
    depth: np.ndarray,
    valid_start: float = 0.2,
    valid_end: float = 1,
    scale: float = 1.0,
) -> np.ndarray:
    """Encode depth array to gray levels.

    Args:
        depth: Depth array
        valid_start: Valid start.
        valid_end: Valid end.
        scale: Scale.

    Returns:
        Depth array in gray levels.
    """
    mask = depth > 1

    # Scale gives depth in mm
    depth_n = depth * scale * 0.001

    if mask.sum() > 0:
        depth_n[mask] -= depth_n[mask].min()
        depth_n[mask] /= depth_n[mask].max() / (valid_end - valid_start)
        depth_n[mask] += valid_start

    depth_n *= 255
    depth_n = cv2.applyColorMap(depth_n[:, :, np.newaxis].astype(np.uint8), cv2.COLORMAP_PLASMA)

    return depth_n


def encode_rle(mask: list[list] | dict, height: int, width: int) -> dict:
    """Encode mask from polygons / uncompressed RLE / RLE to RLE.

    Args:
        mask: Mask as polygons / uncompressed RLE / RLE.
        height: Image height.
        width: Image width.

    Returns:
        Mask as RLE.
    """
    if isinstance(mask, list):
        return polygons_to_rle(mask, height, width)
    elif isinstance(mask, dict):
        if isinstance(mask["counts"], list):
            return urle_to_rle(mask)
        return mask
    raise ValueError("Mask must be a list of polygons or an uncompressed RLE")


def mask_to_rle(mask: Image.Image | np.ndarray) -> dict:
    """Encode mask from Pillow or NumPy array to RLE.

    Args:
        mask: Mask as Pillow or NumPy array.

    Returns:
        Mask as RLE.
    """
    mask_array = np.asfortranarray(mask)
    return mask_api.encode(mask_array)


def rle_to_mask(rle: dict[str, list[int] | bytes]) -> np.ndarray:
    """Decode mask from RLE to NumPy array.

    Args:
        rle: Mask as RLE.

    Returns:
        Mask as NumPy array.
    """
    return mask_api.decode(rle)


def polygons_to_rle(polygons: list[list], height: int, width: int) -> dict:
    """Encode mask from polygons to RLE.

    Args:
        polygons: Mask as polygons.
        height: Image height.
        width: Image width.

    Returns:
        Mask as RLE.
    """
    rles = mask_api.frPyObjects(polygons, height, width)
    return mask_api.merge(rles)


def rle_to_polygons(rle: dict[str, list[int] | bytes]) -> list[list]:
    """Encode mask from RLE to polygons.

    Args:
        rle: Mask as RLE.

    Returns:
        Mask as polygons.
    """
    if "size" not in rle:
        raise ValueError("RLE must have a size")
    h, w = rle["size"]
    polygons, _ = mask_to_polygons(rle_to_mask(rle))

    # Normalize point coordinates
    for p in polygons:
        p[::2] = [x / w for x in p[::2]]
        p[1::2] = [y / h for y in p[1::2]]

    return polygons


def mask_to_polygons(mask: np.ndarray) -> tuple[list[list], bool]:
    """Encode mask from NumPy array to polygons.

    Args:
        mask: Mask as NumPy array

    Returns:
        Tuple:
            - Mask as polygons
            - True if mask has holes
    """
    # Some versions of cv2 does not support incontiguous arr
    mask = np.ascontiguousarray(mask)

    # cv2.RETR_CCOMP flag retrieves all the contours and arranges them
    # to a 2-level hierarchy.
    # External contours (boundary) of the object are placed in hierarchy-1.
    # Internal contours (holes) are placed in hierarchy-2.
    # cv2.CHAIN_APPROX_NONE flag gets vertices of polygons from contours.
    res = cv2.findContours(mask.astype("uint8"), cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)
    hierarchy = res[-1]

    # If mask is empty
    if hierarchy is None:
        return [], False

    # Check if mask has holes
    has_holes = (hierarchy.reshape(-1, 4)[:, 3] >= 0).sum() > 0

    res = res[-2]
    res = [x.flatten() for x in res]

    # The coordinates from OpenCV are integers in range [0, W-1 or H-1].
    # We add 0.5 to turn them into real-value coordinate space. A better solution
    # would be to first +0.5 and then dilate the returned polygon by 0.5.
    res = [x + 0.5 for x in res if len(x) >= 6]

    # Convert np.array to lists
    res = [x.tolist() for x in res]

    return res, has_holes


def urle_to_rle(urle: dict[str, list[int]]) -> dict[str, list[int] | bytes]:
    """Encode mask from uncompressed RLE to RLE.

    Args:
        urle: Mask as uncompressed RLE.

    Returns:
        Mask as RLE.
    """
    height, width = urle["size"]
    return mask_api.frPyObjects(urle, height, width)


def rle_to_urle(rle: dict[str, list[int] | bytes]) -> dict[str, list[int]]:
    """Encode mask from RLE to uncompressed RLE.

    Args:
        rle: Mask as RLE.

    Returns:
        Mask as uncompressed RLE.
    """
    if "counts" not in rle or rle["counts"] is None:
        raise ValueError("RLE must have counts")
    mask = rle_to_mask(rle)
    urle = {"counts": [], "size": list(mask.shape)}

    for i, (value, elements) in enumerate(groupby(mask.ravel(order="F"))):
        urle["counts"].append(0 if i == 0 and value == 1 else len(list(elements)))

    return urle


def mask_area(rle: dict[str, list[int] | bytes]) -> float:
    """Compute mask area.

    Args:
        rle: Mask as RLE

    Returns:
        Mask area
    """
    return float(mask_api.area(rle))


def image_to_base64(image: Image.Image, format: str | None = None) -> str:
    """Encode image from Pillow to base64.

    The image is returned as a base64 string formatted as
    "data:image/{image_format};base64,{base64}".

    Args:
        image: Pillow image.
        format: Image format.

    Returns:
        Image as base64.
    """
    if image.format is None and format is None:
        raise ValueError("Image format is not defined")

    buffered = io.BytesIO()
    out_format = format or image.format
    if out_format.upper() == "UNKNOWN":
        out_format = "JPEG"
    image.save(buffered, format=out_format)

    encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")

    return f"data:image/{out_format.lower()};base64,{encoded}"


def base64_to_image(base64_image: str) -> Image.Image:
    """Decode image from base64 to Pillow.

    Expect the image to be formatted as "data:image/{image_format};base64,{base64}".

    Args:
        base64_image: Image as base64.

    Returns:
        Pillow image.
    """
    image_data = base64.b64decode(base64_image.split(",", maxsplit=1)[1].encode("utf-8"))
    return Image.open(io.BytesIO(image_data))


def get_image_thumbnail(image: Image.Image, size: tuple[int, int]) -> Image.Image:
    """Get image thumbnail.

    Args:
        image: Pillow Image.
        size: Thumbnail size.

    Returns:
        Image thumbnail as Pillow.
    """
    if (
        not isinstance(size, tuple)
        or len(size) != 2
        or not isinstance(size[0], int)
        or not isinstance(size[1], int)
        or size[0] <= 0
        or size[1] <= 0
    ):
        raise ValueError(f"Invalid thumbnail size: {size}")
    thumbnail = image.copy()
    thumbnail.thumbnail(size)
    return thumbnail


def generate_text_image_base64(text: str, width: int = 128, height: int = 128, font_size: int = 16) -> str:
    """Generate a thumbnail displaying given text.

    Args:
        text: input text
        width: thumbnail width
        height: thumbnail height
        font_size: font size

    Returns:
        base64 image of given text.
    """
    image = Image.new("RGB", (width, height), "white")

    try:
        font = ImageFont.truetype("arial.ttf", size=font_size)
    except IOError:
        font = ImageFont.load_default(size=font_size)

    draw = ImageDraw.Draw(image)

    # Set text boundaries
    max_text_width = width - 10  # 5px margin on both sides
    x_start = 5  # Left margin
    y_start = 5  # Top margin
    line_height = font.size + 2  # Line spacing

    # Manually wrap text based on actual pixel width
    words = text.split()
    lines = []
    current_line = ""

    for word in words:
        test_line = current_line + " " + word if current_line else word
        bbox = draw.textbbox((0, 0), test_line, font=font)
        text_width = bbox[2] - bbox[0]

        if text_width <= max_text_width:
            current_line = test_line  # Word fits, add it to the line
        else:
            lines.append(current_line)  # Save current line
            current_line = word  # Start new line with this word

    if current_line:
        lines.append(current_line)  # Add last line

    # Calculate vertical centering
    total_text_height = len(lines) * line_height
    y_position = y_start + (height - total_text_height) // 2

    # Draw text aligned to the left
    for line in lines:
        draw.text((x_start, y_position), line, fill="black", font=font)
        y_position += line_height

    buffered = BytesIO()
    image.save(buffered, format="PNG")
    return "data:image/png;base64," + base64.b64encode(buffered.getvalue()).decode("utf-8")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json
from typing import Any

from pydantic import field_validator

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..base_schema import BaseSchema
from ..entities import Entity
from ..items.item import Item
from ..registry import _register_schema_internal
from ..views.view import View


@_register_schema_internal
class Annotation(BaseSchema):
    """Annotations are used to annotate an entity in a dataset.

    It can refer to an entity, an item, and a view.

    Attributes:
        item_ref: Reference to the annotation's item.
        view_ref: Reference to the annotation's view.
        entity_ref: Reference to the annotation's entity.
        source_ref: Reference to the annotation's source.
        inference_metadata: Metadata of the inference model for the annotation if any.
            dict[str, Any] encoded in a string.
    """

    item_ref: ItemRef = ItemRef.none()
    view_ref: ViewRef = ViewRef.none()
    entity_ref: EntityRef = EntityRef.none()
    source_ref: SourceRef = SourceRef.none()
    inference_metadata: str = json.dumps({})

    @field_validator("inference_metadata", mode="before")
    @classmethod
    def _validate_inference_metadata(cls, v: str | dict[str, Any]) -> str:
        if not isinstance(v, str):
            if not isinstance(v, dict):
                raise ValueError("Metadata must be a dict or string.")
            if not all(isinstance(k, str) for k in v.keys()):
                raise ValueError("Metadata keys must be strings.")
            return json.dumps(v)
        return v

    @property
    def item(self) -> Item:
        """Get the annotation's item."""
        return self.resolve_ref(self.item_ref)

    @property
    def view(self) -> View:
        """Get the annotation's view."""
        return self.resolve_ref(self.view_ref)

    @property
    def entity(self) -> Entity:
        """Get the annotation's entity."""
        return self.resolve_ref(self.entity_ref)


def is_annotation(cls: type, strict: bool = False) -> bool:
    """Check if a class is an Annotation or subclass of Annotation."""
    return issubclass_strict(cls, Annotation, strict)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, field_validator

from pixano.datasets import Dataset
from pixano.features import Item, SchemaGroup, is_item

from .base_schema import BaseSchemaModel
from .table_info import TableInfo


class ItemModel(BaseSchemaModel[Item]):
    """Model for the [Item][pixano.features.Item] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "1",
                    "table_info": {"group": "item", "name": "item", "base_schema": "Item"},
                    "data": {"split": "train", "source": "source1"},
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.ITEM.value:
            raise ValueError(f"Table info group must be {SchemaGroup.ITEM.value}.")
        return value

    def to_row(self, dataset: Dataset) -> Item:
        """Create an [Item][pixano.features.Item] from the model."""
        if not is_item(dataset.schema.schemas[self.table_info.name]):
            raise ValueError(f"Schema type must be a subclass of {Item.__name__}.")
        return super().to_row(dataset)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.workspaces import DefaultVQADatasetItem, WorkspaceType

from .image import ImageFolderBuilder


class VQAFolderBuilder(ImageFolderBuilder):
    """Builder for vqa datasets stored in a folder."""

    WORKSPACE_TYPE = WorkspaceType.IMAGE_VQA
    DEFAULT_SCHEMA: type[DatasetItem] = DefaultVQADatasetItem



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Any, TypeVar

import duckdb
import pandas as pd
import polars as pl
import pyarrow as pa
from lancedb.db import LanceTable
from lancedb.query import LanceEmptyQueryBuilder
from typing_extensions import Self

from pixano.features.schemas.base_schema import BaseSchema


T = TypeVar("T", bound=BaseSchema)


class _PixanoEmptyQueryBuilder(LanceEmptyQueryBuilder):
    def __init__(self, arrow_table):
        self._arrow_table = arrow_table

    def to_arrow(self) -> pa.Table:
        return self._arrow_table


class TableQueryBuilder:
    """Builder class for querying LanceTables.

    It supports the select, where, limit, offset, and order_by clauses:
    - The select clause can be used to select specific columns from the table. If not provided, all columns
        are selected.
    - The where clause can be used to filter the rows of the table.
    - The limit clause can be used to limit the number of rows returned.
    - The offset clause can be used to skip the first n rows.
    - The order_by clause can be used to sort the rows of the table.

    The query is built and executed when calling to_pandas(), to_list(), to_pydantic(), or to_polars().

    Attributes:
        table: The LanceTable to query.
    """

    def __init__(self, table: LanceTable):
        """Initializes the TableQueryBuilder.

        Args:
            table: The LanceTable to query.
        """
        if not isinstance(table, LanceTable):
            raise ValueError("table must be a LanceTable.")

        self.table: LanceTable = table
        self._columns: list[str] | dict[str, str] | None = None
        self._where: str | None = None
        self._limit: int | None = None
        self._offset: int | None = None
        self._order_by: list[str] = []
        self._descending: list[bool] = []
        self._function_called: dict[str, bool] = {
            "select": False,
            "where": False,
            "limit": False,
            "offset": False,
            "order_by": False,
            "build": False,
        }

    def _check_called(self, fn_name: str):
        if self._function_called[fn_name]:
            raise ValueError(f"{fn_name}() can only be called once.")
        elif self._function_called["build"]:
            raise ValueError("build() has already been called.")
        self._function_called[fn_name] = True

    def select(self, columns: str | list[str] | dict[str, str]) -> Self:
        """Selects columns to include in the query.

        Note:
            'id' is always included in the select clause.

        Args:
            columns: The columns to include in the query. If a list, the columns are selected in the order they are
                provided. If a dictionary, the keys are the column names and the values are the aliases.
        """
        self._check_called("select")
        if isinstance(columns, str):
            columns = [columns]

        if isinstance(columns, list) or isinstance(columns, dict):
            if isinstance(columns, list) and not all(isinstance(x, str) for x in columns):
                raise ValueError("columns must be a list of strings.")
            elif isinstance(columns, dict) and not all(
                isinstance(k, str) and isinstance(v, str) for k, v in columns.items()
            ):
                raise ValueError("columns must be a dictionary with string keys and values.")
            if isinstance(columns, list) and "id" not in columns:
                columns = ["id"] + columns
            elif isinstance(columns, dict) and "id" not in columns.values():
                columns["id"] = "id"
            self._columns = columns
        else:
            raise ValueError("columns must be a string, a list of string or a string mapping dictionary.")
        return self

    def where(self, where: str) -> Self:
        """Sets the where clause for the query.

        Args:
            where: The condition to filter the rows.

        Returns:
            The TableQueryBuilder instance.
        """
        self._check_called("where")
        if not isinstance(where, str):
            raise ValueError("where must be a string.")
        self._where = where
        return self

    def limit(self, limit: int | None) -> Self:
        """Sets the limit for the query.

        Args:
            limit: The number of rows to return.

        Returns:
            The TableQueryBuilder instance.
        """
        self._check_called("limit")
        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                raise ValueError("limit must be None or a positive integer.")
        self._limit = limit
        return self

    def offset(self, offset: int | None) -> Self:
        """Sets the offset for the query.

        Args:
            offset: The number of rows to skip.

        Returns:
            The TableQueryBuilder instance.
        """
        self._check_called("offset")
        if offset is not None:
            if not isinstance(offset, int) or offset < 0:
                raise ValueError("offset must be None or a positive integer.")
        self._offset = offset
        return self

    def order_by(self, order_by: str | list[str], descending: bool | list[bool] = False) -> Self:
        """Sets the order_by clause for the query.

        Args:
            order_by: The column(s) to sort by.
            descending: Whether to sort in descending order.

        Returns:
            The TableQueryBuilder instance.
        """
        self._check_called("order_by")
        if isinstance(order_by, str):
            order_by = [order_by]
        elif not isinstance(order_by, list) or not all(isinstance(x, str) for x in order_by):
            raise ValueError("order_by must be a string or a list of strings.")
        if isinstance(descending, bool):
            descending = [descending] * len(order_by)
        elif (
            not isinstance(descending, list)
            or not all(isinstance(x, bool) for x in descending)
            or len(descending) != len(order_by)
        ):
            raise ValueError("descending must be a boolean or a list of booleans with the same length as order_by.")

        self._order_by = order_by
        self._descending = descending
        return self

    def _execute(self) -> pa.Table:
        """Builds the LanceQueryBuilder.

        If order_by or offset are set, the rows are fetched and sorted before building the query.

        Note:
            to_pydantic() and to_list() will call this method internally and keep the order of the rows.

        Returns:
            The LanceQueryBuilder instance if get_order is False, otherwise a tuple with the query and the ordered IDs.
        """
        if all(not self._function_called[fn_name] for fn_name in self._function_called):
            raise ValueError("At least one of select(), where(), limit(), offset(), or order_by() must be called.")
        self._check_called("build")

        if self._columns is None or self._columns == ["*"]:
            # If no columns are selected, select all columns
            # Can have better performance than *
            columns = self.table.schema.names
        else:
            columns = self._columns

        # If order_by
        #   build the query directly using DuckDB
        # Else
        #   build the query using Lance
        # This is because Lance does not support order_by
        # Also DuckDB is less memory efficient than Lance
        if self._order_by != []:
            arrow_table = self.table.to_lance()  # noqa: F841

            def duckdb_format_column(column):
                if "." in column:
                    struct, property = column.split(".")
                    return f"{struct}['{property}']"
                return column

            formatted_columns = [
                duckdb_format_column(col) for col in columns
            ]  # format columns to handle struct columns
            SQL_QUERY = f"SELECT {', '.join(formatted_columns)} FROM arrow_table"
            if self._where is not None:
                SQL_QUERY += f" WHERE {self._where}"
            if self._order_by != []:
                SQL_QUERY += " ORDER BY "
                SQL_QUERY += ", ".join(
                    [f"{col} {'DESC' if desc else 'ASC'}" for col, desc in zip(self._order_by, self._descending)]
                )
            if self._limit is not None:
                SQL_QUERY += f" LIMIT {self._limit}"
            if self._offset is not None:
                SQL_QUERY += f" OFFSET {self._offset}"

            arrow_results: pa.Table = duckdb.query(SQL_QUERY).to_arrow_table()
            arrow_results = arrow_results.rename_columns(columns)  # rename columns to match the requested columns
            return arrow_results
        else:
            limit = self.table.count_rows() if self._limit is None else self._limit
            query = self.table.search(None).select(columns).limit(limit)
            if self._where is not None:
                query = query.where(self._where)
            if self._offset is not None:
                query = query.offset(self._offset)
            return query.to_arrow()

    def to_pandas(self) -> pd.DataFrame:
        """Builds the query and returns the result as a pandas DataFrame.

        Returns:
            The result as a pandas DataFrame.
        """
        return _PixanoEmptyQueryBuilder(self._execute()).to_pandas()

    def to_list(self) -> list[dict[str, Any]]:
        """Builds the query and returns the result as a list of dictionaries.

        Returns:
            The result as a list of dictionaries.
        """
        return _PixanoEmptyQueryBuilder(self._execute()).to_list()

    def to_pydantic(self, model: type[T]) -> list[T]:
        """Builds the query and returns the result as a list of Pydantic models.

        Returns:
            The result as a list of Pydantic models.
        """
        return _PixanoEmptyQueryBuilder(self._execute()).to_pydantic(model)

    def to_polars(self) -> pl.DataFrame:
        """Builds the query and returns the result as a polars DataFrame.

        Returns:
            The result as a polars DataFrame.
        """
        return _PixanoEmptyQueryBuilder(self._execute()).to_polars()



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path

from fastapi import HTTPException
from typing_extensions import TypeVar

from pixano.app.models import AnnotationModel, BaseSchemaModel, EmbeddingModel, EntityModel, ItemModel, ViewModel
from pixano.app.models.sources import SourceModel
from pixano.app.models.table_info import TableInfo
from pixano.app.models.utils import _SCHEMA_GROUP_TO_SCHEMA_MODEL_DICT
from pixano.app.settings import Settings
from pixano.datasets import Dataset
from pixano.datasets.utils import DatasetAccessError, DatasetPaginationError
from pixano.datasets.utils.errors import DatasetIntegrityError
from pixano.features import BaseSchema, SchemaGroup
from pixano.features.schemas.registry import _PIXANO_SCHEMA_REGISTRY
from pixano.utils import get_super_type_from_dict


T = TypeVar("T", bound=BaseSchemaModel)


def get_dataset(dataset_id: str, dir: Path, media_dir: Path | None = None) -> Dataset:
    """Get a dataset.

    If the dataset is not found, raise a 404 error.

    Args:
        dataset_id: Dataset ID.
        dir: Directory containing the dataset.
        media_dir: Directory containing the media files.

    Returns:
        The dataset.
    """
    try:
        dataset = Dataset.find(dataset_id, dir, media_dir)
    except FileNotFoundError:
        raise HTTPException(
            status_code=404,
            detail=f"Dataset {dataset_id} not found in {dir.absolute()}.",
        )
    return dataset


def validate_group(
    group: str | SchemaGroup,
    valid_groups: set[SchemaGroup] = set(SchemaGroup),
) -> SchemaGroup:
    """Assert that a group is valid.

    Args:
        group: Group.
        valid_groups: The valid groups.
    """
    try:
        group = SchemaGroup(group)
    except ValueError:
        raise HTTPException(
            status_code=400,
            detail=f"Group {group} is not a SchemaGroup.",
        )
    if group not in valid_groups:
        raise HTTPException(
            status_code=400,
            detail=f"Group {group.value} is not valid.",
        )
    return group


def assert_table_in_group(dataset: Dataset, table: str, group: SchemaGroup) -> None:
    """Assert that a table belongs to a group.

    If the table does not belong to the group, raise a 404 error.

    Args:
        dataset: Dataset.
        table: Table name.
        group: Group.
    """
    if table in [SchemaGroup.ITEM.value, SchemaGroup.SOURCE.value]:
        return
    elif table not in dataset.schema.groups[group]:
        raise HTTPException(
            status_code=404,
            detail=f"Table {table} is not in the {group.value} group table.",
        )


def get_rows(
    dataset: Dataset,
    table: str,
    where: str | None = None,
    ids: list[str] | None = None,
    item_ids: list[str] | None = None,
    limit: int | None = None,
    skip: int = 0,
) -> list[BaseSchema]:
    """Get rows from a table.

    The rows can be filtered by a where clause, IDs, item IDs or a limit and a skip.

    Args:
        dataset: Dataset containing the table.
        table: Table name.
        where: Where clause.
        ids: IDs of the rows.
        item_ids: Item IDs of the rows.
        limit: Limit number of rows.
        skip: Skip number of rows.

    Returns:
        List of rows.
    """
    try:
        rows = dataset.get_data(table_name=table, where=where, ids=ids, limit=limit, skip=skip, item_ids=item_ids)
    except DatasetPaginationError as err:
        raise HTTPException(status_code=400, detail="Invalid query parameters. " + str(err))
    except DatasetAccessError as err:
        raise HTTPException(status_code=500, detail="Internal server error. " + str(err))

    if rows == [] or rows is None:
        raise HTTPException(
            status_code=404,
            detail=f"No rows found for {dataset.info.id}/{table}.",
        )
    return rows


def get_model_from_row(table: str, model_type: type[T], row: BaseSchema) -> T:
    """Get a model from a row.

    Args:
        table: Table name containing the row.
        model_type: Model type to create.
        row: Row.

    Returns:
        The model.
    """
    try:
        is_group = issubclass(model_type, BaseSchemaModel)
        if not is_group:
            raise HTTPException(status_code=500, detail="Model type is not a subclass of BaseModelSchema.")
    except TypeError:
        raise HTTPException(status_code=500, detail="Model type is not a subclass of BaseModelSchema.")
    if issubclass(model_type, AnnotationModel):
        group = SchemaGroup.ANNOTATION
    elif issubclass(model_type, EmbeddingModel):
        group = SchemaGroup.EMBEDDING
    elif issubclass(model_type, EntityModel):
        group = SchemaGroup.ENTITY
    elif issubclass(model_type, ItemModel):
        group = SchemaGroup.ITEM
    elif issubclass(model_type, SourceModel):
        group = SchemaGroup.SOURCE
    elif issubclass(model_type, ViewModel):
        group = SchemaGroup.VIEW
    else:
        raise HTTPException(
            status_code=500,
            detail="Model type not correct.",
        )

    pixano_schema_type = get_super_type_from_dict(type(row), _PIXANO_SCHEMA_REGISTRY)

    if pixano_schema_type is None:
        raise HTTPException(
            status_code=500,
            detail="Schema type not found in registry.",
        )
    table_info = TableInfo(name=table, group=group.value, base_schema=pixano_schema_type.__name__)
    model = model_type.from_row(row, table_info)
    return model


def get_models_from_rows(
    table: str,
    model_type: type[T],
    rows: list[BaseSchema],
) -> list[T]:
    """Get models from rows.

    Args:
        table: Table name containing the rows.
        model_type: Model type to create.
        rows: Rows.

    Returns:
        List of models.
    """
    return [get_model_from_row(table, model_type, row) for row in rows]


def delete_rows(
    dataset: Dataset,
    table: str,
    ids: list[str],
) -> list[str]:
    """Delete rows from a table.

    Args:
        dataset: Dataset containing the table.
        table: Table name.
        ids: IDs of the rows to delete.

    Returns:
        IDs not found.
    """
    try:
        ids_not_found = dataset.delete_data(table, ids)
    except ValueError:
        raise HTTPException(
            status_code=400,
            detail="Invalid query parameters. ids and item_ids cannot be set at the same time",
        )
    return ids_not_found


def update_rows(
    dataset: Dataset,
    table: str,
    models: list[BaseSchemaModel],
) -> list[BaseSchema]:
    """Update rows in a table.

    Args:
        dataset: Dataset containing the table.
        table: Table name.
        models: Models of the rows to update.

    Returns:
        The updated rows.
    """
    try:
        rows: list[BaseSchema] = BaseSchemaModel.to_rows(models, dataset)
    except Exception as err:
        raise HTTPException(
            status_code=400,
            detail="Invalid data.\n" + str(err),
        )

    try:
        updated_rows = dataset.update_data(table, rows)
    except DatasetIntegrityError as err:
        raise HTTPException(status_code=400, detail="Dataset integrity compromised.\n" + str(err))
    except ValueError as err:
        raise HTTPException(
            status_code=400,
            detail="Invalid data.\n" + str(err),
        )

    # TODO: return updated rows instead of input rows
    # TODO: check if rows are updated or created which change HTTP status code
    return updated_rows


def create_rows(
    dataset: Dataset,
    table: str,
    models: list[BaseSchemaModel],
) -> list[BaseSchema]:
    """Add rows to a table.

    Args:
        dataset: Dataset containing the table.
        table: Table name.
        models: Models of the rows to add.

    Returns:
        The added rows.
    """
    try:
        rows: list[BaseSchema] = BaseSchemaModel.to_rows(models, dataset)
    except Exception as err:
        raise HTTPException(
            status_code=400,
            detail="Invalid data.\n" + str(err),
        )

    try:
        created_rows = dataset.add_data(table, rows)
    except DatasetIntegrityError as err:
        raise HTTPException(status_code=400, detail="Dataset integrity compromised.\n" + str(err))
    except ValueError as err:
        raise HTTPException(
            status_code=400,
            detail="Invalid data.\n" + str(err),
        )

    return created_rows


async def get_rows_handler(
    dataset_id: str,
    group: SchemaGroup,
    table: str,
    settings: Settings,
    where: str | None = None,
    ids: list[str] | None = None,
    item_ids: list[str] | None = None,
    limit: int | None = None,
    skip: int = 0,
) -> list[BaseSchemaModel]:
    """Get row models.

    Rows can be filtered by a where clause, IDs, item IDs or a limit and a skip.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name.
        settings: App settings.
        where: Where clause.
        ids: IDs of the rows.
        item_ids: Item IDs of the rows.
        limit: Limit number of rows.
        skip: Skip number of rows.

    Returns:
        List of models.
    """
    group = validate_group(group)
    dataset = get_dataset(dataset_id, settings.library_dir, settings.media_dir)
    assert_table_in_group(dataset, table, group)
    rows = get_rows(dataset=dataset, table=table, where=where, ids=ids, item_ids=item_ids, limit=limit, skip=skip)
    models = get_models_from_rows(table, _SCHEMA_GROUP_TO_SCHEMA_MODEL_DICT[group], rows)
    return models


async def get_row_handler(
    dataset_id: str, group: SchemaGroup, table: str, id: str, settings: Settings
) -> BaseSchemaModel:
    """Get a row model.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name.
        id: ID of the row.
        settings: App settings.

    Returns:
        The model.
    """
    return (await get_rows_handler(dataset_id, group, table, settings, ids=[id], item_ids=None, limit=None, skip=0))[0]


async def create_rows_handler(
    dataset_id: str,
    group: SchemaGroup,
    table: str,
    rows: list[BaseSchemaModel],
    settings: Settings,
) -> list[BaseSchemaModel]:
    """Add rows to a table.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name containing the rows.
        rows: Rows to add.
        settings: App settings.

    Returns:
        List of updated rows.
    """
    group = validate_group(group)
    dataset = get_dataset(dataset_id, settings.library_dir, settings.media_dir)
    assert_table_in_group(dataset, table, group)
    rows_rows = create_rows(dataset, table, rows)
    rows_models = get_models_from_rows(table, _SCHEMA_GROUP_TO_SCHEMA_MODEL_DICT[group], rows_rows)
    return rows_models


async def create_row_handler(
    dataset_id: str,
    group: SchemaGroup,
    table: str,
    id: str,
    row: BaseSchemaModel,
    settings: Settings,
) -> BaseSchemaModel:
    """Add a row to a table.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name.
        id: ID of the row.
        row: Row to add.
        settings: App settings.

    Returns:
        The added row.
    """
    if id != row.id:
        raise HTTPException(status_code=400, detail="ID in path and body do not match.")
    return (await create_rows_handler(dataset_id=dataset_id, group=group, table=table, rows=[row], settings=settings))[
        0
    ]


async def update_row_handler(
    dataset_id: str,
    group: SchemaGroup,
    table: str,
    id: str,
    row: BaseSchemaModel,
    settings: Settings,
) -> BaseSchemaModel:
    """Update a row in a table.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name containing the row.
        id: ID of the row.
        row: Row to update.
        settings: App settings.

    Returns:
        The updated row.
    """
    if id != row.id:
        raise HTTPException(status_code=400, detail="ID in path and body do not match.")
    return (await update_rows_handler(dataset_id=dataset_id, group=group, table=table, rows=[row], settings=settings))[
        0
    ]


async def update_rows_handler(
    dataset_id: str,
    group: SchemaGroup,
    table: str,
    rows: list[BaseSchemaModel],
    settings: Settings,
) -> list[BaseSchemaModel]:
    """Update rows in a table.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name containing the rows.
        rows: Rows to update.
        settings: App settings.

    Returns:
        List of updated rows.
    """
    group = validate_group(group)
    dataset = get_dataset(dataset_id, settings.library_dir, settings.media_dir)
    assert_table_in_group(dataset, table, group)
    row_rows = update_rows(dataset, table, rows)
    row_models = get_models_from_rows(table, _SCHEMA_GROUP_TO_SCHEMA_MODEL_DICT[group], row_rows)
    return row_models


async def delete_row_handler(dataset_id: str, group: SchemaGroup, table: str, id: str, settings: Settings) -> None:
    """Delete a row from a table.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name containing the row.
        id: ID of the row to delete.
        settings: App settings.
    """
    return await delete_rows_handler(dataset_id, group, table, ids=[id], settings=settings)


async def delete_rows_handler(
    dataset_id: str,
    group: SchemaGroup,
    table: str,
    ids: list[str],
    settings: Settings,
) -> None:
    """Delete rows from a table.

    Args:
        dataset_id: Dataset ID containing the table.
        group: Schema group of the schema of the table.
        table: Table name containing the rows.
        ids: IDs of the rows to delete.
        settings: App settings.
    """
    group = validate_group(group)
    dataset = get_dataset(dataset_id, settings.library_dir, settings.media_dir)
    assert_table_in_group(dataset, table, group)
    delete_rows(dataset, table, ids)
    return None



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.app.main import create_app
from pixano.app.serve import App


__all__ = [
    "App",
    "create_app",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from datetime import datetime

from .base_type import BaseType
from .nd_array_float import NDArrayFloat


_ATOMIC_PYTHON_TYPES: list[type] = [
    int,
    float,
    complex,
    str,
    bool,
    bytes,
    bytearray,
    memoryview,
]

_TYPES_REGISTRY: dict[str, type] = {"BaseType": BaseType, "datetime": datetime, "NDArrayFloat": NDArrayFloat}
_PIXANO_TYPES_REGISTRY: dict[str, type[BaseType]] = _TYPES_REGISTRY.copy()


def _add_type_to_registry(cls, registry: dict[str, type[BaseType]]) -> None:
    if not (cls in _ATOMIC_PYTHON_TYPES or issubclass(cls, BaseType)):
        raise ValueError(f"Table type {type} must be a an atomic python type or derive from BaseType.")

    cls_name = cls.__name__
    if cls_name in registry:
        raise ValueError(f"Type {cls} already registered")
    registry[cls_name] = cls
    return None


def _register_type_internal(cls):
    _add_type_to_registry(cls, _TYPES_REGISTRY)
    if issubclass(cls, BaseType):
        _add_type_to_registry(cls, _PIXANO_TYPES_REGISTRY)
    return cls


for python_type in _ATOMIC_PYTHON_TYPES:
    _register_type_internal(python_type)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import logging
from abc import ABC, abstractmethod
from datetime import timedelta
from pathlib import Path
from typing import Any, Iterator, Literal

import lancedb
import shortuuid
import tqdm
from lancedb.table import Table

from pixano.datasets import Dataset, DatasetFeaturesValues, DatasetInfo, DatasetItem, DatasetSchema
from pixano.datasets.utils.integrity import check_dataset_integrity, handle_integrity_errors
from pixano.features import BaseSchema, Item, SchemaGroup
from pixano.features.schemas.source import Source, SourceKind


logger = logging.getLogger(__name__)


class DatasetBuilder(ABC):
    """Abstract base class for dataset builders.

    To build a dataset, inherit from this class, implement the `generate_data` method and launch the `build` method.

    Attributes:
        target_dir: The target directory for the dataset.
        previews_path: The path to the previews directory.
        info: Dataset information (name, description, ...).
        dataset_schema: The schema of the dataset.
        schemas: The schemas of the dataset tables.
        db: The connection to the `LanceDB` database of the dataset.
    """

    def __init__(
        self,
        target_dir: Path | str,
        dataset_item: type[DatasetItem],
        info: DatasetInfo,
    ):
        """Initialize a DatasetBuilder instance.

        Args:
            target_dir: The target directory for the dataset.
            dataset_item: The dataset item schema.
            info: Dataset information (name, description, ...).
        """
        self.target_dir: Path = Path(target_dir)
        self.previews_path: Path = self.target_dir / Dataset._PREVIEWS_PATH

        self.info: DatasetInfo = info
        self.dataset_schema: DatasetSchema = dataset_item.to_dataset_schema()
        self.schemas: dict[str, type[BaseSchema]] = self.dataset_schema.schemas

        self.db: lancedb.DBConnection = lancedb.connect(self.target_dir / Dataset._DB_PATH)

    @property
    def item_schema(self) -> type[Item]:
        """The item schema for the dataset."""
        return self.dataset_schema.schemas[SchemaGroup.ITEM.value]

    @property
    def item_schema_name(self) -> str:
        """The item schema name for the dataset."""
        return SchemaGroup.ITEM.value

    def add_source(
        self,
        name: str,
        kind: str | SourceKind,
        metadata: str | dict[str, Any] = {},
        id: str = "",
    ) -> str:
        """Add a source to the dataset.

        Args:
            name: Name of the source.
            kind: Kind of source.
            metadata: Metadata of the source. If a dict is provided, it is converted to a JSON string.
            id: The id of the source. If not provided, a random id is generated.

        Returns:
            The id of the source.
        """
        if id == "":
            id = shortuuid.uuid()
        if isinstance(kind, str):
            kind = SourceKind(kind)
        self.db.open_table("source").add([Source(id=id, name=name, kind=kind.value, metadata=metadata)])
        return id

    def add_ground_truth_source(self, metadata: str | dict[str, Any] = {}) -> str:
        """Add a ground truth source to the dataset.

        Args:
            metadata: Metadata of the ground truth source.

        Returns:
            The id of the ground truth source.
        """
        return self.add_source(
            id=SourceKind.GROUND_TRUTH.value, name="Ground Truth", kind=SourceKind.GROUND_TRUTH, metadata=metadata
        )

    def build(
        self,
        mode: Literal["add", "create", "overwrite"] = "create",
        flush_every_n_samples: int | None = None,
        compact_every_n_transactions: int | None = None,
        check_integrity: Literal["raise", "warn", "none"] = "raise",
    ) -> Dataset:
        """Build the dataset.

        It generates data from the source directory and insert them in the tables of the database.

        Args:
            mode: The mode for creating the tables in the database.
                The mode can be "create", "overwrite" or "add":
                    - "create": Create the tables in the database. If the tables already exist, an error is raised.
                    - "overwrite": Overwrite the tables in the database.
                    - "add": Append to the tables in the database.
            flush_every_n_samples: The number of samples accumulated from `generate_data` before they are
                flushed in tables. The counter is per table. If None, data are inserted at each iteration.
            compact_every_n_transactions: The number of transactions before compacting each table.
                If None, the dataset is compacted only at the end.
            check_integrity: The integrity check to perform after building the dataset. It can be "raise",
                "warn" or "none":
                    - "raise": Raise an error if integrity errors are found.
                    - "warn": Print a warning if integrity errors are found.
                    - "none": Do not check integrity.

        Returns:
            The built dataset.
        """
        if mode not in ["add", "create", "overwrite"]:
            raise ValueError(f"mode should be 'add', 'create' or 'overwrite' but got {mode}")
        if check_integrity not in ["raise", "warn", "none"]:
            raise ValueError(f"check_integrity should be 'raise', 'warn' or 'none' but got {check_integrity}")
        if flush_every_n_samples is not None and flush_every_n_samples <= 0:
            raise ValueError(f"flush_every_n_samples should be greater than 0 but got {flush_every_n_samples}")
        if compact_every_n_transactions is not None and compact_every_n_transactions <= 0:
            raise ValueError(
                f"compact_every_n_transactions should be greater than 0 but got {compact_every_n_transactions}"
            )

        # save info.json
        self.info.id = shortuuid.uuid() if self.info.id == "" else self.info.id
        self.info.to_json(self.target_dir / Dataset._INFO_FILE)

        # save features_values.json
        # TMP: empty now
        DatasetFeaturesValues().to_json(self.target_dir / Dataset._FEATURES_VALUES_FILE)

        # remove previous schema.json if any
        if (self.target_dir / Dataset._SCHEMA_FILE).exists():
            (self.target_dir / Dataset._SCHEMA_FILE).unlink()
        # save schema.json
        self.dataset_schema.to_json(self.target_dir / Dataset._SCHEMA_FILE)

        if mode == "add":
            tables = self.open_tables()
        else:
            tables = self.create_tables(mode)

        # accumulate items to insert in tables
        accumulate_data_tables: dict[str, list] = {table_name: [] for table_name in tables.keys()}
        # count transactions per table
        transactions_per_table: dict[str, int] = dict.fromkeys(tables.keys(), 0)

        logger.info(f"Building dataset {self.info.name}")
        for items in tqdm.tqdm(self.generate_data(), desc=f"Generate data for dataset {self.info.name}"):
            # assert that items have keys that are in tables
            for table_name, item_value in items.items():
                if item_value is None or item_value == []:
                    continue
                if table_name not in tables:
                    raise KeyError(f"Table {table_name} not found in tables")

                accumulate_data_tables[table_name].extend(item_value if isinstance(item_value, list) else [item_value])

                # make transaction every n iterations per table
                if len(accumulate_data_tables[table_name]) > 0 and (
                    flush_every_n_samples is None
                    or len(accumulate_data_tables[table_name]) % flush_every_n_samples == 0
                ):
                    table = tables[table_name]
                    table.add(accumulate_data_tables[table_name])
                    transactions_per_table[table_name] += 1
                    accumulate_data_tables[table_name] = []

                # compact dataset every n transactions per table
                if (
                    compact_every_n_transactions is not None
                    and transactions_per_table[table_name] % compact_every_n_transactions == 0
                    and transactions_per_table[table_name] > 0
                ):
                    self.compact_table(table_name)

        # make transaction for final batch
        for table_name, table in tables.items():
            if len(accumulate_data_tables[table_name]) > 0:
                table.add(accumulate_data_tables[table_name])
        self.compact_dataset()

        logger.info(f"Dataset {self.info.name} built in {self.target_dir} with id {self.info.id}")

        dataset = Dataset(self.target_dir)

        if check_integrity != "none":
            logger.info(f"Checking dataset {dataset.info.name} integrity...")
            handle_integrity_errors(check_dataset_integrity(dataset), raise_or_warn=check_integrity)

        logger.info(f"Dataset {dataset.info.name} built successfully.")
        return dataset

    def compact_table(self, table_name: str) -> None:
        """Compact a table by cleaning up old versions and compacting files.

        Args:
            table_name: The name of the table to compact.
        """
        table = self.db.open_table(table_name)
        table.compact_files(
            target_rows_per_fragment=1048576, max_rows_per_group=1024, materialize_deletions=False, num_threads=None
        )
        table.cleanup_old_versions(older_than=timedelta(days=0), delete_unverified=True)

    def compact_dataset(self) -> None:
        """Compact the dataset by calling `compact_table` for each table in the database."""
        for table_name in self.schemas.keys():
            self.compact_table(table_name)

    @abstractmethod
    def generate_data(self) -> Iterator[dict[str, BaseSchema | list[BaseSchema]]]:
        """Generate data from the source directory.

        It should yield a dictionary with keys corresponding to the table names and values corresponding to the data.

        It must be implemented in the subclass.

        Returns:
            An iterator over the data following the dataset schemas.
        """
        raise NotImplementedError

    def create_tables(
        self,
        mode: Literal["create", "overwrite"] = "create",
    ) -> dict[str, Table]:
        """Create tables in the database.

        Returns:
            The tables in the database.
        """
        tables = {}
        for key, schema in self.schemas.items():
            self.db.create_table(key, schema=schema, mode=mode)

            tables[key] = self.db.open_table(key)
        self.db.create_table("source", schema=Source, mode=mode)
        tables["source"] = self.db.open_table("source")

        return tables

    def open_tables(self) -> dict[str, Table]:
        """Open the tables in the database.

        Returns:
            The tables in the database.
        """
        tables = {}
        for key in self.schemas.keys():
            tables[key] = self.db.open_table(key)
        tables["source"] = self.db.open_table("source")

        return tables



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, field_validator
from typing_extensions import TypeVar

from pixano.datasets import Dataset
from pixano.features import Entity, SchemaGroup, is_entity

from .base_schema import BaseSchemaModel
from .table_info import TableInfo


T = TypeVar("T", bound=Entity)


class EntityModel(BaseSchemaModel[Entity]):
    """Model for the [Entity][pixano.features.Entity] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "cat_n1",
                    "table_info": {"group": "entities", "name": "cats", "base_schema": "Entity"},
                    "data": {
                        "item_ref": {"name": "item", "id": "1"},
                        "view_ref": {"name": "image", "id": "orange_cats"},
                        "parent_ref": {"name": "", "id": ""},
                        "category": "orange",
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.ENTITY.value:
            raise ValueError(f"Table info group must be {SchemaGroup.ENTITY.value}.")
        return value

    def to_row(self, dataset: Dataset) -> Entity:
        """Create an [Entity][pixano.features.Entity] from the model."""
        if not is_entity(dataset.schema.schemas[self.table_info.name]):
            raise ValueError(f"Schema type must be a subclass of {Entity.__name__}.")
        return super().to_row(dataset)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Body, Depends
from pixano_inference.pydantic import ModelConfig
from pixano_inference.pydantic.models import ModelInfo

from pixano.app.routers.inference.utils import get_client_from_settings
from pixano.app.settings import Settings, get_settings


router = APIRouter(prefix="/models", tags=["Models"])


@router.get("/list/", response_model=list[ModelInfo])
async def list_models(
    settings: Annotated[Settings, Depends(get_settings)], task: str | None = None
) -> list[ModelInfo]:
    """List all models from pixano inference client."""
    client = get_client_from_settings(settings)
    models = await client.list_models()
    if task is None:
        return models
    return [m for m in models if task == m.task]


@router.post("/instantiate")
async def instantiate_model(
    config: Annotated[ModelConfig, Body(embed=True)],
    provider: Annotated[str, Body(embed=True)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Instantiate a model from pixano inference client."""
    client = get_client_from_settings(settings)
    return await client.instantiate_model(provider=provider, config=config)


@router.delete("/delete/{model_name}")
async def delete_model(
    model_name: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete a model from pixano inference client."""
    client = get_client_from_settings(settings)
    return await client.delete_model(model_name=model_name)
